# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

# pylint: disable=invalid-name, unused-argument, no-value-for-parameter
# ruff: noqa: N802, N815, E501

from zapf.device import typecode_description
from zapf.simulator.funcs import adr, memcpy, memset, sizeof
from zapf.simulator.runtime import (
    Globals,
    Struct,
    Var,
    array,
    boolean,
    byte,
    dword,
    lreal,
    lword,
    program,
    real,
    string,
    word,
)
from zapf.spec import FLOAT32_MAX
from zapf.spec.v_2021_09 import DEVICE_FLAGS, encode_unit

PILS_MAGIC = 2021.09
DESCRIPTOR_SLOT_SIZE = 48
DESCRIPTOR_SLOTS = 4
INDEXER_DATA_SIZE = DESCRIPTOR_SLOTS * DESCRIPTOR_SLOT_SIZE
INDEXER_SIZE = 2 + INDEXER_DATA_SIZE
INDEXER_OFFSET = 6
MSGIO_NDATA = 22
PARAM_DEFAULTS = [0, 100, 10, 90, 5, 500, 0, 10, 2, 0, 0.125, 0, 10, 1, 0, 0, 0, 0]


class DevicesLayout(Struct):
    # sdi64/sdo64/kv64
    sdx64_value = Var(lword, 0)
    sdx64_target = Var(lword, 0)

    # sai64/sao64/rv64
    sax64_value = Var(lreal, 0)
    sax64_target = Var(lreal, 0)

    # sdi32/sdo32/kv32
    sdx32_value = Var(dword, 0)
    sdx32_target = Var(dword, 0)

    # sai32/sao32/rv32
    sax32_value = Var(real, 0)
    sax32_target = Var(real, 0)

    # sdi16/sdo16/kv16
    sdx16_value = Var(word, 0)
    sdx16_target = Var(word, 0)

    # sw16/sw32
    estatus = Var(dword, 0x20000000)

    # 1e04 = discrete output with extended status
    dox_value = Var(word, 0)
    dox_target = Var(word, 0)
    dox_estatus = Var(dword, 0)

    # msgio (0x050c)
    msg_mbox = Var(word, 0)
    msg_data = Var(array(byte, 0, MSGIO_NDATA-1))

    # di64, do64 (re-using do64_target as di64_value)
    dx64_value = Var(lword, 0)
    dx64_target = Var(lword, 0)
    dx64_estatus = Var(dword, 0x10000000)
    dx64_nerrid = Var(word, 0)
    dx64_reserved = Var(word, 0)

    # di32, do32  (re-using do32_target as di32_value)
    dx32_value = Var(dword, 0)
    dx32_target = Var(dword, 0)
    dx32_estatus = Var(dword, 0x10000000)
    dx32_padding = Var(dword, 0)

    # di16, do16  (re-using do16_target as di16_value)
    dx16_value = Var(word, 0)
    dx16_target = Var(word, 0)
    dx16_status = Var(word, 0x1000)
    dx16_padding = Var(word, 0)

    # ai64, ao64  (re-using ao64_target as ai64_value)
    ax64_value = Var(lreal, 0)
    ax64_target = Var(lreal, 0)
    ax64_estatus = Var(dword, 0x10000000)
    ax64_nerrid = Var(word, 0)
    ax64_reserved = Var(word, 0)

    # ai32, ao32  (re-using ao32_target as ai32_value)
    ax32_value = Var(real, 0)
    ax32_target = Var(real, 0)
    ax32_estatus = Var(dword, 0x10001000)
    ax32_padding = Var(dword, 0)

    #  flatin/out64
    fax64_value = Var(lreal, 50)
    fax64_target = Var(lreal, 50)
    fax64_estatus = Var(dword, 0x10000000)
    fax64_nerrid = Var(word, 0)
    fax64_reserved = Var(word, 0)
    fax64_params = Var(array(lreal, 0, 17), PARAM_DEFAULTS)

    # flatin/out32
    fax32_value = Var(real, 50)
    fax32_target = Var(real, 50)
    fax32_estatus = Var(dword, 0x10001000)
    fax32_params = Var(array(real, 0, 17), PARAM_DEFAULTS)
    fax32_padding = Var(dword, 0)

    # pi/po64
    px64_value = Var(lreal, 50)
    px64_target = Var(lreal, 0)
    px64_estatus = Var(dword, 0x10000000)
    px64_nerrid = Var(word, 0)
    px64_pctl = Var(word, 0)
    px64_pvalue = Var(lreal, 0)

    # pi/po32
    px32_value = Var(real, 50)
    px32_target = Var(real, 0)
    px32_estatus = Var(dword, 0x10000000)
    px32_nerrid = Var(word, 0)
    px32_pctl = Var(word, 0)
    px32_pvalue = Var(real, 0)

    # legacy pi/po32
    lpx32_value = Var(real, 50)
    lpx32_target = Var(real, 0)
    lpx32_status = Var(word, 0x1000)
    lpx32_pctl = Var(word, 0)
    lpx32_pvalue = Var(real, 0)

    # v2i/v2o32
    v2x32_values = Var(array(real, 0, 1))
    v2x32_targets = Var(array(real, 0, 1))
    v2x32_estatus = Var(dword, 0x10000000)
    v2x32_nerrid = Var(word, 0)
    v2x32_pctl = Var(word, 0)
    v2x32_pvalue = Var(real, 0)

    # legacy v2i/v2o32
    lv2x32_values = Var(array(real, 0, 1))
    lv2x32_targets = Var(array(real, 0, 1))
    lv2x32_status = Var(word, 0x1000)
    lv2x32_pctl = Var(word, 0)
    lv2x32_pvalue = Var(real, 0)

    # v2i/v2o64
    v2x64_values = Var(array(lreal, 0, 1))
    v2x64_targets = Var(array(lreal, 0, 1))
    v2x64_estatus = Var(dword, 0x10000000)
    v2x64_nerrid = Var(word, 0)
    v2x64_pctl = Var(word, 0)
    v2x64_pvalue = Var(lreal, 0)

    # v16i/v16o64
    v16x64_values = Var(array(lreal, 0, 15))
    v16x64_targets = Var(array(lreal, 0, 15))
    v16x64_estatus = Var(dword, 0x10000000)
    v16x64_nerrid = Var(word, 0)
    v16x64_pctl = Var(word, 0)
    v16x64_pvalue = Var(lreal, 0)

    # v16i/v16o32
    v16x32_values = Var(array(real, 0, 15))
    v16x32_targets = Var(array(real, 0, 15))
    v16x32_estatus = Var(dword, 0x10000000)
    v16x32_nerrid = Var(word, 0)
    v16x32_pctl = Var(word, 0)
    v16x32_pvalue = Var(real, 0)

    # legacy v16i/v16o32
    lv16x32_values = Var(array(real, 0, 15))
    lv16x32_targets = Var(array(real, 0, 15))
    lv16x32_status = Var(word, 0x1000)
    lv16x32_pctl = Var(word, 0)
    lv16x32_pvalue = Var(real, 0)

    # table storage: 1d
    table64_act_row = Var(word, 0)
    table64_req_row = Var(word, 0)
    table64_line = Var(array(lreal, 0, 9))
    # table storage: 2d, direct mapped
    table64 = Var(array(lreal, 0, 99))

    # table storage: 1d
    table32_act_row = Var(word, 0)
    table32_req_row = Var(word, 0)
    table32_line = Var(array(real, 0, 9))
    # table storage: 2d, direct mapped
    table32 = Var(array(real, 0, 99))

    rv32 = Var(real, 0)
    rv64 = Var(lreal, 0)
    kw64 = Var(lword, 0)
    kw32 = Var(dword, 0)
    kw16 = Var(word, 0)
    padding = Var(word, 0)

    lax32_value = Var(real, 0)
    lax32_target = Var(real, 0)
    lax32_status = Var(word, 0x1000)
    lax32_padding = Var(word, 0)

    lfax32_value = Var(real, 50)
    lfax32_target = Var(real, 50)
    lfax32_status = Var(word, 0x1000)
    lfax32_padding = Var(word, 0)
    lfax32_params = Var(array(real, 0, 17), PARAM_DEFAULTS)


# ensure byte addresses above are accurate
assert DevicesLayout.sizeof() == 2600

alignment_ok = True
for _n, addr in DevicesLayout.OFFSET.items():
    o = getattr(DevicesLayout, _n)
    try:
        size = o.dtype.INNER.sizeof()
    except AttributeError:
        size = o.dtype.sizeof()
    if addr != (addr // size) * size:
        print(f'{_n} is not correctly aligned! (addr={addr}, size={size})')
        alignment_ok = False
assert alignment_ok


def addrof(n):
    return 200 + DevicesLayout.OFFSET[n]


class PLCDescriptor(Struct):
    dt = Var(word, 0x1010)
    devices = Var(word, 0)
    description = Var(word, 0)
    version = Var(word, 0)
    author = Var(word, 0)
    desc_slot_size = Var(word, DESCRIPTOR_SLOT_SIZE)
    num_devices = Var(word, 0)
    flags = Var(word, (DESCRIPTOR_SLOTS - 1) << 12)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 16 - 1), '')
    # flags, desc_slot_size == const, num_devices -> magic!
    keys = 'dt', 'devices', 'description', 'version', 'author', 'name'


class DeviceDescriptor(Struct):
    dt = Var(word, 0x2014)
    prev = Var(word, 0)
    description = Var(word, 0)
    value_param = Var(word, 0)
    aux = Var(word, 0)
    params = Var(word, 0)
    errid = Var(word, 0)
    typecode = Var(word, 0)
    address = Var(word, 0)
    flags = Var(word, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 20 - 1), '')
    keys = 'dt', 'prev', 'description', 'value_param', 'aux', 'params', \
        'errid', 'typecode', 'address', 'flags', 'name'


class StringDescriptor(Struct):
    dt = Var(word, 0x3004)
    prev = Var(word, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 4 - 1), '')
    keys = 'dt', 'prev', 'name'


class EnumDescriptor(Struct):
    dt = Var(word, 0x4006)
    prev = Var(word, 0)
    # don't name a field 'value' or 'addr' or the simulator may ignore writes to it...
    val = Var(word, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 6 - 1), '')
    keys = 'dt', 'prev', 'val', 'name'


class BitfieldDescriptor(Struct):
    dt = Var(word, 0x5008)
    prev = Var(word, 0)
    enum_id = Var(word, 0)
    lsb = Var(byte, 0)
    width = Var(byte, 1)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 8 - 1), '')
    keys = 'dt', 'prev', 'enum_id', 'lsb', 'width', 'name'


class FlagDescriptor(Struct):
    dt = Var(word, 0x5105)
    prev = Var(word, 0)
    lsb = Var(byte, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 5 - 1), '')
    keys = 'dt', 'prev', 'lsb', 'name'


class NumericParameterDescriptor(Struct):
    dt = Var(word, 0x6114)
    prev = Var(word, 0)
    description = Var(word, 0)
    paramidx = Var(word, 0)
    paramtype = Var(word, 0)
    unit = Var(word, 0)
    minval = Var(real, -FLOAT32_MAX)
    maxval = Var(real, FLOAT32_MAX)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 20 - 1), '')
    keys = 'dt', 'prev', 'description', 'paramidx', 'paramtype', 'unit', \
        'minval', 'maxval', 'name'


class EnumParameterDescriptor(Struct):
    dt = Var(word, 0x620E)
    prev = Var(word, 0)
    description = Var(word, 0)
    read_enum_id = Var(word, 0)
    write_enum_id = Var(word, 0)
    paramidx = Var(word, 0)
    paramtype = Var(word, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 14 - 1), '')
    keys = 'dt', 'prev', 'description', 'read_enum_id', 'write_enum_id', \
        'paramidx', 'paramtype', 'name'


class SpecialFuncDescriptor(Struct):
    dt = Var(word, 0x680E)
    prev = Var(word, 0)
    description = Var(word, 0)
    arg_id = Var(word, 0)
    res_id = Var(word, 0)
    paramidx = Var(word, 0)
    paramtype = Var(word, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 14 - 1), '')
    keys = 'dt', 'prev', 'description', 'arg_id', 'res_id', 'paramidx', \
        'paramtype', 'name'


class TableDescriptor(Struct):
    dt = Var(word, 0x6C10)
    prev = Var(word, 0)
    description = Var(word, 0)
    columns = Var(word, 0)
    base = Var(word, 0)
    flags = Var(word, 0)
    last_row = Var(word, 0)
    last_column = Var(word, 0)
    name = Var(string(DESCRIPTOR_SLOT_SIZE - 16 - 1), '')
    keys = 'dt', 'prev', 'description', 'columns', 'base', 'flags', \
        'last_row', 'last_column', 'name'


class DebugDescriptor(Struct):
    dt = Var(word, 0xFF06)
    cycle = Var(word, 0)
    indexer_size = Var(word, 0)
    text = Var(string(INDEXER_DATA_SIZE - 6 - 1), '')  # Yes: this is BIG
    keys = 'dt', 'cycle', 'text'


class ST_Indexer(Struct):  # noqa: N801
    Request = Var(word, 0)
    #    Data = Var(array(U_Descriptor, 0, DESCRIPTOR_SLOTS-1))
    Data = Var(array(array(byte, 0, DESCRIPTOR_SLOT_SIZE - 1), 0, DESCRIPTOR_SLOTS - 1))


# singletons


PLC_DESCRIPTOR = PLCDescriptor(name='testplc_2021_09.py').alloc_self()
DEBUG_DESCRIPTOR = DebugDescriptor(text='PLC_Problem 42, please call 137',
                                   indexer_size=INDEXER_DATA_SIZE).alloc_self()


# map descriptorid/descriptortuple to descriptor
DESCRIPTOR_ARRAY = [PLC_DESCRIPTOR]
descriptors = {}


def add_descr(t, descr):
    # pylint: disable=global-statement
    if t in descriptors:
        return descriptors[t]
    next_descriptor = len(DESCRIPTOR_ARRAY)
    kwds = dict(zip(descr.keys, t))
    d = descr(**kwds).alloc_self()
    descriptors[t] = next_descriptor
    DESCRIPTOR_ARRAY.append(d)
    return next_descriptor


def add_Device(name, typecode, address, description='', value_param=None,
               aux=0, params=0, errid=0, flags=0, unit=''):
    if value_param is None:
        # provide some sensible default for most devices
        if typecode in (0x1201, 0x1401, 0x1602, 0x1A02, 0x1E03, 0x1E04):
            value_param = add_IntParam(0, '', 'rw', 0, -32768, 32767)
        elif typecode in (0x1202, 0x1402, 0x1604, 0x1A04, 0x1E06):
            value_param = add_IntParam(0, '', 'rw', 0, -(2 ** 31), 2 ** 31 - 1)
        elif typecode in (0x1204, 0x1404, 0x1608, 0x1A08, 0x1E0C):
            value_param = add_IntParam(0, '', 'rw', 0, -(2 ** 63), 2 ** 63 - 1)
        elif typecode >> 8 != 5:  # not for MSGIO !
            value_param = add_FloatParam(0, '', 'rw', 0, -FLOAT32_MAX, FLOAT32_MAX,
                                         encode_unit(unit) if isinstance(unit, str) else unit)
        else:
            value_param = 0
    if isinstance(flags, str):
        flags = [flags]
    try:
        int(flags)
    except TypeError:
        f = 0
        for bit, fn in DEVICE_FLAGS.items():
            if fn in flags:
                f |= 1 << bit
        flags = f
    t = (
        0x2014,
        PLC_DESCRIPTOR.devices,
        add_String(description),
        value_param,
        aux,
        params,
        errid,
        typecode,
        address,
        flags,
        name,
    )
    PLC_DESCRIPTOR.devices = add_descr(t, DeviceDescriptor)
    PLC_DESCRIPTOR.num_devices += 1
    # no return....


def add_String(s):
    prev = 0
    while s:
        part, s = s[: DESCRIPTOR_SLOT_SIZE - 4 - 1], s[DESCRIPTOR_SLOT_SIZE - 4 - 1:]
        t = (0x3004, prev, part)
        prev = add_descr(t, StringDescriptor)
    return prev


def add_Enum(prev, val, name):
    t = (0x4006, prev, val, name)
    return add_descr(t, EnumDescriptor)


def add_Enums(prev, enumdict):
    # add all values of a dict to the given chain.
    # XXX: also remove duplicates aready in prev? How?
    for v, n in sorted(enumdict.items()):
        prev = add_Enum(prev, v, n)
    return prev


def add_Bitfield(prev, name, lsb, width, enum):
    if isinstance(enum, dict):
        enum = add_Enums(0, enum)
    t = (0x5008, prev, enum, lsb, width, name)
    return add_descr(t, BitfieldDescriptor)


def add_Flag(prev, lsb, name):
    t = (0x5105, prev, lsb, name)
    return add_descr(t, FlagDescriptor)


def add_BfFChain(prev, chain):
    # chain is an iterable of 2/4 tuples for flags/bitfields
    # flags: (bit, name)
    # bitfield: (name, lsb, width, enumdict)
    for e in chain:
        if len(e) == 2:
            prev = add_Flag(prev, *e)
        elif len(e) == 4:
            prev = add_Bitfield(prev, *e)
        else:
            raise RuntimeError(f'bad Element {e} in Bitfield/Flag Chain: {chain}')
    return prev


WRO_MAP = {'rw': 1, 'ro': 2, 'obs': 3}  # writeable, readonly, observable


def add_IntParam(prev, name, wro, idx, minval, maxval, unit=0, description=''):
    if isinstance(description, str):
        description = add_String(description)
    wro = WRO_MAP.get(wro, wro) & 3
    if isinstance(unit, str):
        unit = encode_unit(unit)
    t = (0x6114, prev, description, idx, (0xC + wro) << 12, unit, minval, maxval, name)
    return add_descr(t, NumericParameterDescriptor)


def add_FloatParam(prev, name, wro, idx, minval, maxval, unit='', description=''):
    if isinstance(description, str):
        description = add_String(description)
    wro = WRO_MAP.get(wro, wro) & 3
    if isinstance(unit, str):
        unit = encode_unit(unit)
    t = (0x6114, prev, description, idx, (0x4 + wro) << 12, unit, minval, maxval, name)
    return add_descr(t, NumericParameterDescriptor)


def add_EnumParam(prev, name, wro, idx, read_enum, write_enum=0, description=''):
    if isinstance(description, str):
        description = add_String(description)
    wro = WRO_MAP.get(wro, wro) & 3
    if read_enum == 0:
        read_enum = write_enum
    if isinstance(read_enum, dict):
        if isinstance(write_enum, dict):
            read_enum.update(write_enum)
        read_enum = add_Enums(0, read_enum)
    if isinstance(write_enum, dict):
        write_enum = add_Enums(0, write_enum)
    t = (0x620E, prev, description, read_enum, write_enum, idx, (0x8 + wro) << 12, name)
    return add_descr(t, EnumParameterDescriptor)


def add_SFunc(prev, name, idx, arg_id, res_id, description=''):
    if isinstance(description, str):
        description = add_String(description)
    t = (0x680E, prev, description, arg_id, res_id, idx, 0, name)
    return add_descr(t, SpecialFuncDescriptor)


def add_Table(prev, name, base, typ, columns, last_row, last_column, description=''):
    if isinstance(description, str):
        description = add_String(description)
    t = (0x6C10, prev, description, columns, base, (typ & 3) << 14, last_row, last_column, name)
    return add_descr(t, TableDescriptor)


# prepare some chains to be used later

AUX8 = add_BfFChain(0, [(d, f'AUX{d}') for d in range(6)] +
                    [('Minion', 6, 2, {1: 'Ape', 2: 'Banana', 3: 'Dragon'})])

AUX24 = add_BfFChain(AUX8, [(d, f'AUX{d}') for d in range(8, 20)] +
                     [('Minion', 6, 3, {4: 'Hamster', 2:''})] +
                     [(d, f'AUX{d}') for d in range(20, 24)] +
                     [('', 21, 1, {1:'aux22', 0:'nix'})]+
                     [('X', 18, 3, {})] +
                     [('', 22, 2, {0:'', 1:'Under voltage', 2:'Over voltage', 3:'no Power'})])

ERRID = add_BfFChain(
    0, [('lowerByte', 0, 8, 0), ('topmost 7 Bit', 9, 7, 0), (8, 'a Flag')],
)

PARS_1 = add_FloatParam(0, 'UserMin', 'rw', 0, 0, 100, '#', 'lower user settable limit')
PARS_2 = add_FloatParam(PARS_1, 'UserMax', 'rw', 1, 0, 100, '#', 'upper user settable limit')
PARS_3 = add_FloatParam(PARS_2, 'WarnMin', 'rw', 2, 0, 100, '#', 'lower warn limit')
PARS_4 = add_FloatParam(PARS_3, 'WarnMax', 'rw', 3, 0, 100, '#', 'upper warn limit')
PARS_5 = add_FloatParam(PARS_4, 'Timeout', 'rw', 4, 0, 900, 's', 'timeout for movement in s')
PARS_6 = add_FloatParam(PARS_5, 'MaxTravelDist', 'rw', 5, 0, 500, '#', 'maximum travel distance')
PARS_7 = add_FloatParam(PARS_6, 'Offset', 'ro', 6, 0, 100, '#', 'internal offset')
PARS_8 = add_FloatParam(PARS_7, 'P', 'rw', 7, 0, 100, '%/(#)', 'P constant for regulation')
PARS_9 = add_FloatParam(PARS_8, 'I', 'rw', 8, 0, 100, 's', 'I constant for regulation')
PARS_10 = add_FloatParam(PARS_9, 'D', 'rw', 9, 0, 100, '1/s', 'D constant for regulation')
PARS_11 = add_FloatParam(PARS_10, 'Hysteresis', 'rw', 10, 0, 100, '#', 'Hysteresis for regulation')
PARS_12 = add_FloatParam(PARS_11, 'Holdback', 'rw', 11, 0, 100, '#', 'max difference between actual temp and setpoint')
PARS_13 = add_FloatParam(PARS_12, 'Speed', 'rw', 12, 0.1, 100, '#/s', 'max speed of movement')
PARS_14 = add_FloatParam(PARS_13, 'Accel', 'rw', 13, 0, 100, '#/s^2', 'acceleration of movement')
PARS_15 = add_FloatParam(PARS_14, 'Home', 'rw', 14, 0, 100, '#', 'Home Position')
PARS_16 = add_FloatParam(PARS_15, 'Setpoint', 'obs', 15, 0, 100, '#', 'actual setpoint')
PARS_17 = add_EnumParam(PARS_16, 'microsteps', 'rw', 16, 0, {s:f'{2**s} steps' for s in range(9)}, 'microstepping selection')
PARS_18 = add_IntParam(PARS_17, 'numbits', 'rw', 17, 1, 32, 'bit', 'number of bits per ssi transfer')

TCOL_ = add_EnumParam(PARS_10, 'enable', 'rw', 0, {2:'enabled', 1:'updating', 0:'ignored'}, {2:'enabled', 0:'ignored'},
                      'en/disable this table row')
TCOL = add_EnumParam(TCOL_, 'microsteps', 'rw', 9, 0, {s:f'{2**s} steps' for s in range(9)}, 'microstepping selection')

SFUNC_1 = add_SFunc(PARS_18, 'reset_to_factory_default', 128,
                    add_IntParam(0, 'unlock_value', 'rw', 0, -2**31, 2**31-1, '', 'Unlock value'),
                    0,
                    'resets everything to factory defaults',
                    )
SFUNC_2 = add_SFunc(SFUNC_1, 'home', 133, 0, 0, 'starts a homing cycle')
SFUNC_3 = add_SFunc(SFUNC_2, 'SetPosition', 137, PARS_15, 0, 'sets a new current position')
SFUNC_4 = add_SFunc(SFUNC_3, 'ContMove', 142, PARS_13, 0, 'Starts a continuous movement')
SFUNC_5 = add_SFunc(SFUNC_4, 'SetBits', 240, PARS_18, PARS_18, 'set and read-back current number of bits')
SFUNC_6 = add_SFunc(SFUNC_5, 'GetBits', 241, 0, PARS_18, 'read-back current number of bits')

TABLE_2 = add_Table(SFUNC_6, 'table2', addrof('table32'), 2, TCOL, 9, 9,
                    '32 bit extra simple 10x10 table, direct mapped')
TABLE_1 = add_Table(TABLE_2, 'table1', addrof('table32_act_row'), 1, TCOL, 9, 9,
                    '32 bit extra simple 10x10 table, with line select')
TABLE32 = add_Table(TABLE_1, 'table0', 1000, 0, TCOL, 9, 9,
                    '32 bit extra simple 10x10 table, via paramctlif')

TABLE_3 = add_Table(SFUNC_6, 'table2', addrof('table64'), 2, TCOL, 9, 9,
                    '64 bit extra simple 10x10 table, direct mapped')
TABLE_4 = add_Table(TABLE_3, 'table1', addrof('table64_act_row'), 1, TCOL, 9, 9,
                    '64 bit extra simple 10x10 table, with line select')
TABLE64 = add_Table(TABLE_4, 'table0', 1000, 0, TCOL, 9, 9,
                    '64 bit extra simple 10x10 table, via paramctlif')

ENUM_1 = add_Enums(0, {0: 'On', 1: 'Off'})
ENUM_2 = add_Enum(ENUM_1, 2, 'Moving')

# for valuetype
VT_ENUM = add_EnumParam(0, 'ignored pname', 'rw', -1, ENUM_2, ENUM_1, 'ignored description')
VT_SINT16 = add_IntParam(0, 'ignored pname', 'rw', -1, -32768, 32767, 'Ohm', 'signed 16 bit integer')
VT_UINT16 = add_IntParam(0, 'ignored pname', 'rw', -1, 0, 65535, 'bit', 'unsigned 16 bit integer')
VT_SINT32 = add_IntParam(0, 'ignored pname', 'rw', -1, -2**31, 2**31-1, 'counts', 'signed 32 bit integer')
VT_UINT32 = add_IntParam(0, 'ignored pname', 'rw', -1, 0, 2**32-1, '', 'unsigned 32 bit integer')
VT_SINT64 = add_IntParam(0, 'ignored pname', 'rw', -1, -2**63, 2**63-1, '', 'signed 64 bit integer')
VT_UINT64 = add_IntParam(0, 'ignored pname', 'rw', -1, 0, 2**64-1, '', 'unsigned 64 bit integer')
VT_FLOAT32 = add_FloatParam(0, 'ignored pname', 'rw', -1, -FLOAT32_MAX, FLOAT32_MAX, 'cm', '32 bit float')
VT_FLOAT64 = add_FloatParam(0, 'ignored pname', 'rw', -1, -FLOAT32_MAX, FLOAT32_MAX, 'km', '64 bit float')
VT_V2 = add_FloatParam(VT_FLOAT32, 'ignored channel2', 'rw', 42, 0, 137, '%', 'ignored description of second channel')
VT_V3 = add_FloatParam(VT_V2, 'ignored channel3', 'rw', 49, 0, 100, 'kH/min', 'ignored description of 3rd channel')
VT_V4 = add_FloatParam(VT_V3, 'ignored channel4', 'rw', 50, 0, FLOAT32_MAX, '#', 'ignored description of 4th channel')
VT_V5 = add_FloatParam(VT_V4, 'ignored channel5', 'rw', 51, 0, FLOAT32_MAX, 'm', 'ignored description of 5th channel')
VT_V6 = add_FloatParam(VT_V5, 'ignored channel6', 'rw', 52, 0, FLOAT32_MAX, 'm', 'ignored description of 6th channel')
VT_V7 = add_FloatParam(VT_V6, 'ignored channel7', 'rw', 53, 0, FLOAT32_MAX, 'm', 'ignored description of 7th channel')
VT_V8 = add_FloatParam(VT_V7, 'ignored channel8', 'rw', 54, 0, FLOAT32_MAX, 'm', 'ignored description of 8th channel')
VT_V16 = VT_V8
for j in range(9, 17):
    VT_V16 = add_FloatParam(VT_V16, f'ignored channel {j}', 'ro', 60 + j, 0, j * j * j * j * j * j - 1, j,
                            f'ignored description {j}')

# create devices
# note: Name fields will be overriden later in Init().
# here they are only a programming/ers hint.
add_Device('sdi16', 0x1201, addrof('sdx16_value'), value_param=VT_ENUM, description='simple discrete input, 16 bit')
add_Device('sdi32', 0x1202, addrof('sdx32_value'), value_param=VT_SINT32, description='simple discrete input, 32 bit',
           params=TABLE_1)
add_Device('sdi64', 0x1204, addrof('sdx64_value'), value_param=VT_SINT64, description='simple discrete input, 64 bit',
           params=TABLE_4)

add_Device('sai32', 0x1302, addrof('sax32_value'), value_param=VT_FLOAT32, description='simple analog input, 32 bit')
add_Device('sai64', 0x1304, addrof('sax64_value'), value_param=VT_FLOAT64, description='simple analog input, 64 bit')

add_Device('kw16', 0x1401, addrof('kw16'), value_param=VT_UINT16, description='key word, 16 bit')
add_Device('kw32', 0x1402, addrof('kw32'), value_param=VT_UINT32, description='key word, 32 bit')
add_Device('kw64', 0x1404, addrof('kw64'), value_param=VT_UINT64, description='key word, 64 bit')

add_Device('rv32', 0x1502, addrof('rv32'), value_param=VT_FLOAT32, description='real value, 32 bit')
add_Device('rv64', 0x1504, addrof('rv64'), value_param=VT_FLOAT64, description='real value, 64 bit')

add_Device('sdo16', 0x1602, addrof('sdx16_value'), value_param=VT_SINT16, description='simple discrete output, 16 bit')
add_Device('sdo32', 0x1604, addrof('sdx32_value'), value_param=VT_SINT32, description='simple discrete output, 32 bit')
add_Device('sdo64', 0x1608, addrof('sdx64_value'), value_param=VT_SINT64, description='simple discrete output, 64 bit')

add_Device('sao32', 0x1704, addrof('sax32_value'), value_param=VT_FLOAT32, description='simple analog output, 32 bit')
add_Device('sao64', 0x1708, addrof('sax64_value'), value_param=VT_FLOAT64, description='simple analog output, 64 bit')

add_Device('sw16', 0x1801, addrof('estatus')+2, value_param=VT_UINT16, description='status word, 16 bit',
           aux=AUX8)
add_Device('sw32', 0x1802, addrof('estatus'), value_param=VT_UINT32, description='''\
status word, 32 bit

This is a longer description which surely will use more than a single descriptor to store.
it is in 'git' format, i.e. a single line with short 'title', an empty line, followed by a longer
explanation. rst formatting may be used (suggested by Beckhoff).

The device is a special device, reflecting an extended statusword.
There is no functionality behind it (in this implementation).

Let's see how this really long description (with some special utf-8 chars)
is displayed on various clients....

;;; ⁂⁇
''', aux=AUX24)  # noqa: RUF001

add_Device('di16', 0x1a02, addrof('dx16_target'), value_param=VT_UINT16, description=None,
           aux=AUX8)
add_Device('di32', 0x1a04, addrof('dx32_target'), value_param=VT_UINT32, description='discrete input, 32 bit',
           aux=AUX24)
add_Device('di64', 0x1a08, addrof('dx64_target'), value_param=VT_UINT64, description='discrete input, 64 bit',
           aux=AUX24, errid=ERRID)

add_Device('lai32', 0x1b03, addrof('lax32_target'), value_param=VT_FLOAT32, description='legacy analog input, 32 bit',
           aux=AUX8)
add_Device('ai32', 0x1b04, addrof('ax32_target'), unit='kA', description='analog input, 32 bit',
           aux=AUX24)
add_Device('ai64', 0x1b08, addrof('ax64_target'), unit='Mcounts/d', description='analog input, 64 bit',
           aux=AUX24, errid=ERRID)

add_Device('ldo16', 0x1e03, addrof('dx16_value'), value_param=VT_ENUM, description='legacy discrete output, 16 bit',
           aux=AUX8)
add_Device('do16', 0x1e04, addrof('dox_value'), value_param=VT_SINT16, description='discrete output, 16 bit',
           aux=AUX24)
add_Device('do32', 0x1e06, addrof('dx32_value'), value_param=VT_SINT32, description='discrete output, 32 bit',
           aux=AUX24)
add_Device('do64', 0x1e0c, addrof('dx64_value'), value_param=VT_SINT64, description='discrete output, 64 bit',
           aux=AUX24, errid=ERRID)

add_Device('lao32', 0x1f05, addrof('lax32_value'), value_param=VT_FLOAT32, description='legacy analog output, 32 bit',
           aux=AUX8)
add_Device('ao32', 0x1f06, addrof('ax32_value'), unit='kA', description='analog output, 32 bit',
           aux=AUX24)
add_Device('ao64', 0x1f0c, addrof('ax64_value'), value_param=VT_FLOAT64, description='analog output, 64 bit',
           aux=AUX24, errid=ERRID)

add_Device('lfa32i1', 0x2006, addrof('lfax32_target'), unit='mm', description='legacy flat analog input, 32 bit, 1 param',
           aux=AUX8, params=PARS_1)
add_Device('lfa32i16', 0x2f24, addrof('lfax32_target'), unit='mm', description='legacy flat analog input, 32 bit, 16 params',
           aux=AUX8, params=PARS_16, errid=ERRID)

add_Device('fa32i1', 0x6006, addrof('fax32_target'), unit='mm', description='flat analog input, 32 bit, 1 param',
           aux=AUX8, params=PARS_1)
add_Device('fa32i16', 0x6f24, addrof('fax32_target'), unit='mm', description='flat analog input, 32 bit, 16 params',
           aux=AUX8, params=PARS_16, errid=ERRID)

add_Device('fa64i1', 0x200c, addrof('fax64_target'), unit='mm', description='flat analog input, 64 bit, 1 param',
           aux=AUX24, params=PARS_1)
add_Device('fa64i16', 0x2f48, addrof('fax64_target'), unit='mm', description='flat analog input, 64 bit, 16 params',
           aux=AUX24, params=PARS_16, errid=ERRID)

add_Device('lfa32o1', 0x3008, addrof('lfax32_value'), unit='mm', description='legacy flat analog output, 32 bit, 1 param',
           aux=AUX8, params=PARS_1)
add_Device('lfa32o16', 0x3f26, addrof('lfax32_value'), unit='mm', description='legacy flat analog output, 32 bit, 16 params',
           aux=AUX8, params=PARS_16, errid=ERRID)

add_Device('fa32o1', 0x7008, addrof('fax32_value'), unit='mm', description='flat analog output, 32 bit, 1 param',
           aux=AUX8, params=PARS_1)
add_Device('fa32o16', 0x7f26, addrof('fax32_value'), unit='mm', description='flat analog output, 32 bit, 16 params',
           aux=AUX8, params=PARS_16, errid=ERRID)

add_Device('fa64o1', 0x3010, addrof('fax64_value'), unit='mm', description='flat analog output, 64 bit, 1 param',
           aux=AUX24, params=PARS_1)
add_Device('fa64o16', 0x3f4c, addrof('fax64_value'), unit='mm', description='flat analog output, 64 bit, 16 params',
           aux=AUX24, params=PARS_16,errid=ERRID)

add_Device('lpi32', 0x4006, addrof('lpx32_target'), unit='%', description='legacy parameter input, 32 bit',
           aux=AUX8, params=SFUNC_6, errid=ERRID)
add_Device('pi32', 0x4008, addrof('px32_target'), unit='%', description='parameter output, 32 bit',
           aux=AUX24, params=TABLE32, errid=ERRID)
add_Device('lpo32', 0x5008, addrof('lpx32_value'), value_param=0, description='legacy parameter input, 32 bit',
           aux=AUX8, params=SFUNC_6, errid=ERRID)
add_Device('po32', 0x500a, addrof('px32_value'), unit='%', description='parameter output, 32 bit',
           aux=AUX24, errid=ERRID, params=TABLE32)

add_Device('pi64', 0x400c, addrof('px64_target'), unit='%', description='parameter input, 64 bit',
           aux=AUX24, params=TABLE64, errid=ERRID)
add_Device('po64', 0x5010, addrof('px64_value'), unit='%', description='parameter output, 64 bit',
           aux=AUX24, params=TABLE64, errid=ERRID)

add_Device('lv2i32', 0x4108, addrof('lv2x32_targets'), value_param=VT_V2, description='legacy vector 2 input, 32 bit',
           aux=AUX8, params=SFUNC_6)
add_Device('lv16i32', 0x4f24, addrof('lv16x32_targets'), value_param=VT_V16, description='legacy vector 16 input, 32 bit',
           aux=AUX8, params=SFUNC_6)

add_Device('v2i32', 0x410a, addrof('v2x32_targets'), value_param=VT_V2, description='vector 2 input, 32 bit',
           aux=AUX24, params=SFUNC_6)
add_Device('v16i32', 0x4f26, addrof('v16x32_targets'), value_param=VT_V16, description='vector 16 input, 32 bit',
           aux=AUX24, params=SFUNC_6)

add_Device('lv2o32', 0x510c, addrof('lv2x32_values'), value_param=VT_V2, description='legacy vector 2 output, 32 bit',
           aux=AUX8, params=SFUNC_6)
add_Device('lv16o32', 0x5f44, addrof('lv16x32_values'), value_param=VT_V16, description='legacy vector 16 output, 32 bit',
           aux=AUX8, params=SFUNC_6)

add_Device('v2o32', 0x510e, addrof('v2x32_values'), value_param=VT_V2, description='vector 2 output, 32 bit',
           aux=AUX24, params=SFUNC_6)
add_Device('v16o32', 0x5f46, addrof('v16x32_values'), value_param=VT_V16, description='vector 16 output, 32 bit',
           aux=AUX24, params=SFUNC_6)

add_Device('v2i64', 0x4110, addrof('v2x64_targets'), value_param=VT_V2, description='vector 2 input, 64 bit',
           aux=AUX24, params=SFUNC_6)
add_Device('v16i64', 0x4f48, addrof('v16x64_targets'), value_param=VT_V16, description='vector 16 input, 64 bit',
           aux=AUX24, params=SFUNC_6)

add_Device('v2o64', 0x5118, addrof('v2x64_values'), value_param=VT_V2, description='vector 2 output, 64 bit',
           aux=AUX24, params=SFUNC_6)
add_Device('v16o64', 0x5f88, addrof('v16x64_values'), value_param=VT_V16, description='vector 16 output, 64 bit',
           aux=AUX24, params=SFUNC_6)

add_Device('msg22', 0x0501 + MSGIO_NDATA//2, addrof('msg_mbox'), description='MessageIO with 22 bytes data area')

add_Device('lpi32e', 0x4006, addrof('lpx32_target'), value_param=VT_ENUM, description='legacy parameter input, 32 bit with enums',
           aux=AUX8, params=SFUNC_6, errid=ERRID)
add_Device('pi32e', 0x4008, addrof('px32_target'), value_param=VT_ENUM, description='parameter output, 32 bit with enums',
           aux=AUX24, params=TABLE32, errid=ERRID)
add_Device('lpo32e', 0x5008, addrof('lpx32_value'), value_param=VT_ENUM, description='legacy parameter input, 32 bit with enums',
           aux=AUX8, params=SFUNC_6, errid=ERRID)
add_Device('po32e', 0x500a, addrof('px32_value'), value_param=VT_ENUM, description='parameter output, 32 bit with enums',
           aux=AUX24, errid=ERRID, params=TABLE32)


PLC_DESCRIPTOR.description = add_String('simulation for testing zapf')
PLC_DESCRIPTOR.version = add_String('https://forge.frm2.tum.de/review/mlz/pils/zapf:v2.1-alpha')
PLC_DESCRIPTOR.author = add_String('anonymous\nauthor')


class Global(Globals):
    fMagic = Var(real, PILS_MAGIC, at='%MB0')
    iOffset = Var(word, INDEXER_OFFSET, at='%MB4')

    stIndexer = Var(ST_Indexer, at=f'%MB{INDEXER_OFFSET}')

    data = Var(DevicesLayout, at='%MB200')

    iCycle = Var(word, 0)


g = Global()

@program(
    nDevices = Var(word),
    devnum = Var(word),
    infotype = Var(word),
    itemp = Var(byte),
    tempofs = Var(word),
)
def Indexer(_v):
    if g.fMagic != PILS_MAGIC:
        g.fMagic = PILS_MAGIC
    if g.iOffset != INDEXER_OFFSET:
        g.iOffset = INDEXER_OFFSET

    g.iCycle += 1

    req_num = g.stIndexer.Request & 0x7FFF

    data = g.stIndexer.Data

    memset(adr(data), 0, sizeof(data))

    if req_num == 32767:
        DEBUG_DESCRIPTOR.cycle = g.iCycle
        memcpy(adr(data), adr(DEBUG_DESCRIPTOR), sizeof(DEBUG_DESCRIPTOR))
    else:
        for i in range(DESCRIPTOR_SLOTS):
            if req_num + i < len(DESCRIPTOR_ARRAY):
                memcpy(adr(data) + i*DESCRIPTOR_SLOT_SIZE,
                       adr(DESCRIPTOR_ARRAY[req_num + i]),
                       DESCRIPTOR_SLOT_SIZE)

    g.stIndexer.Request[[15]] = 1


# helper
def advance32(value, target, status, usermin=0.0, usermax=100.0,
              warnmin=10.0, warnmax=90.0, speed=10.0):
    state = status >> 12
    reason = 0
    if state == 0:
        state = 1
    elif state in [1, 3, 7]:
        state = 1 if warnmin <= value <= warnmax else 3
    elif state == 5:
        state, reason = (6, 0) if usermin <= target <= usermax else (8, 1)
    elif state == 6:
        if value == target:
            state, reason = 1, 0
        value += max(min((target - value), speed / 20), -speed / 20)
    target = max(min(usermax, target), usermin)
    if value < warnmin:
        reason |= 4
    if value > warnmax:
        reason |= 8

    return value, target, (state << 12) | (reason << 8) | (1 << ((g.iCycle >> 9) & 7))


# same as above but for 32 bit status fields
def advance64(value, target, status, *args, **kwds):
    v, t, s = advance32(value, target, max(status >> 16, status & 0xFFFF), *args, **kwds)
    return v, t, s << 16 | s


# pylint: disable=too-many-return-statements
def handle_pctl(pctl, pvalue, valuestore, tablestore):
    pnum = pctl & 8191
    cmd = pctl >> 13
    if pctl & 0x8000:
        # nothing to do
        return pctl, pvalue
    if not cmd:
        return 0x2000, pvalue
    if pnum in [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16,
                17, 128, 133, 137, 142, 240, 241] or 1000 <= pnum <= 1099:
        # param exists, can at least be read
        if cmd == 1:  # DO_READ
            if 0 <= pnum <= 17:
                # return DONE and value
                return 0x8000 | pnum, valuestore[pnum]
            if pnum == 240:
                return 0x8000 | pnum, valuestore[17]
            if pnum == 241:
                return 0x8000 | pnum, valuestore[17]
            if 1000 <= pnum <= 1099:
                # table access
                # return DONE and value
                return 0x8000 | pnum, float(tablestore[pnum - 1000])
            # invent value and DONE
            return 0x8000 | pnum, 0
        if cmd == 2:  # DO_WRITE
            if 0 <= pnum <= 17:
                valuestore[pnum] = pvalue
                return 0x2000 | pnum, pvalue  # -> read
            if pnum == 240:
                valuestore[17] = pvalue
                return 0x8000 | pnum, valuestore[17]
            if pnum == 241:
                return 0x8000 | pnum, valuestore[17]
            if 128 <= pnum <= 239:
                # go to BUSY for some and ERR_RETRY for others
                if pnum in (128, 133):
                    return 0x6000 | pnum, pvalue
                return 0xE000 | pnum, pvalue  # ERR_RETRY
            if 1000 <= pnum <= 1099:
                tablestore[pnum - 1000] = float(pvalue)
                return 0x2000 | pnum, pvalue  # -> read
            # no function, no storage -> ERR_READ_ONLY
            return 0xC000 | pnum, pvalue
        if cmd == 3:  # BUSY
            return 0x2000 | pnum, pvalue
        raise RuntimeError(f'handle_pctl({pctl}, {pvalue}, {valuestore}): '
                           'This should never happen!')
    # non exisiting param -> ERR_NO_IDX
    return 0xA000 | pnum, pvalue


def handle_sxo(value, target, step):
    """Handle the simple a/d outputs."""
    if value < target:
        return min(value + step, target)
    if value > target:
        return max(value - step, target)
    return value


def handle_dx16(value, target, status):
    if status >> 12 == 5:
        return target, 0x6000
    return value, 0x1000


def handle_dx32(value, target, status):
    v, s = handle_dx16(value, target, max(status >> 16, status & 0xFFFF))
    return v, s << 16 | s


@program(msg=Var(array(byte, 0, 80)), msglen=Var(byte, 0))
def Implementation(v):
    d = g.data

    if g.iCycle & 0x3FF == 0:
        d.kw16 = 1 << (g.iCycle >> 12)
        d.kw32 = 1 << (g.iCycle >> 11)
        d.kw64 = 1 << (g.iCycle >> 10)

    d.sax64_value = handle_sxo(d.sax64_value, d.sax64_target, 0.12)
    d.sax32_value = handle_sxo(d.sax32_value, d.sax32_target, 0.12)

    d.sdx64_value = handle_sxo(d.sdx64_value, d.sdx64_target, 1)
    d.sdx32_value = handle_sxo(d.sdx32_value, d.sdx32_target, 1)
    d.sdx16_value = handle_sxo(d.sdx16_value, d.sdx16_target, 1)

    d.ax64_value, d.ax64_target, d.ax64_estatus = advance64(
        d.ax64_value, d.ax64_target, d.ax64_estatus,
    )
    d.ax64_nerrid = d.ax64_estatus >> 15 if d.ax64_estatus[[31]] else 0

    d.ax32_value, d.ax32_target, d.ax32_estatus = advance64(
        d.ax32_value, d.ax32_target, d.ax32_estatus,
    )
    d.lax32_value, d.lax32_target, d.lax32_status = advance32(
        d.lax32_value, d.lax32_target, d.lax32_status,
    )

    d.dx16_target &= 0x3FFF
    d.dx16_value, d.dx16_status = handle_dx16(
        d.dx16_value, d.dx16_target, d.dx16_status,
    )

    d.dx32_target &= 0xFFFFFF
    d.dx32_value, d.dx32_estatus = handle_dx32(
        d.dx32_value, d.dx32_target, d.dx32_estatus,
    )

    d.dx64_target &= 0xFFFFFFFF
    d.dx64_value, d.dx64_estatus = handle_dx32(
        d.dx64_value, d.dx64_target, d.dx64_estatus,
    )
    d.dx64_nerrid = d.dx16_value

    d.dox_value, d.dox_estatus = handle_dx32(d.dox_value, d.dox_target, d.dox_estatus)

    for i in range(16):
        d.fax64_params[i] = max(0, d.fax64_params[i])
    d.fax64_value, d.fax64_target, d.fax64_estatus = advance64(
        d.fax64_value, d.fax64_target, d.fax64_estatus,
        usermin=d.fax64_params[0], usermax=d.fax64_params[1],
        warnmin=d.fax64_params[2], warnmax=d.fax64_params[3],
        speed=d.fax64_params[12],
    )
    d.fax64_nerrid = d.fax64_estatus >> 15 if d.fax64_estatus[[31]] else 0

    for i in range(16):
        d.fax32_params[i] = max(0, d.fax32_params[i])
    d.fax32_value, d.fax32_target, d.fax32_estatus = advance64(
        d.fax32_value, d.fax32_target, d.fax32_estatus,
        usermin=d.fax32_params[0], usermax=d.fax32_params[1],
        warnmin=d.fax32_params[2], warnmax=d.fax32_params[3],
        speed=d.fax32_params[12],
    )

    for i in range(16):
        d.lfax32_params[i] = max(0, d.lfax32_params[i])
    d.lfax32_value, d.lfax32_target, d.lfax32_status = advance32(
        d.lfax32_value, d.lfax32_target, d.lfax32_status,
        usermin=d.lfax32_params[0], usermax=d.lfax32_params[1],
        warnmin=d.lfax32_params[2], warnmax=d.lfax32_params[3],
        speed=d.lfax32_params[12],
    )

    # pi/po share params with fax
    d.px64_value, d.px64_target, d.px64_estatus = advance64(
        d.px64_value, d.px64_target, d.px64_estatus,
        usermin = d.fax64_params[0], usermax = d.fax64_params[1],
        warnmin = d.fax64_params[2], warnmax = d.fax64_params[3],
        speed = d.fax64_params[12],
    )
    d.px64_nerrid = d.px64_estatus >> 15 if d.px64_estatus[[31]] else 0
    d.px64_pctl, d.px64_pvalue = handle_pctl(
        d.px64_pctl, d.px64_pvalue, d.fax64_params, d.table64,
    )

    d.px32_value, d.px32_target, d.px32_estatus = advance64(
        d.px32_value, d.px32_target, d.px32_estatus,
        usermin = d.fax32_params[0], usermax = d.fax32_params[1],
        warnmin = d.fax32_params[2], warnmax = d.fax32_params[3],
        speed = d.fax32_params[12],
    )
    d.px32_nerrid = d.px32_estatus >> 15 if d.px32_estatus[[31]] else 0
    d.px32_pctl, d.px32_pvalue = handle_pctl(
        d.px32_pctl, d.px32_pvalue, d.fax32_params, d.table32,
    )

    d.lpx32_value, d.lpx32_target, d.lpx32_status = advance32(
        d.lpx32_value, d.lpx32_target, d.lpx32_status,
        usermin = d.fax32_params[0], usermax = d.fax32_params[1],
        warnmin = d.fax32_params[2], warnmax = d.fax32_params[3],
        speed = d.fax32_params[12],
    )
    d.lpx32_pctl, d.lpx32_pvalue = handle_pctl(
        d.lpx32_pctl, d.lpx32_pvalue, d.fax32_params, d.table32,
    )

    # v*x devices are pain
    maxstat = 0
    for i in range(2):
        d.v2x64_values[i], d.v2x64_targets[i], stat = advance64(
            d.v2x64_values[i], d.v2x64_targets[i], d.v2x64_estatus,
            usermin=d.fax64_params[0], usermax=d.fax64_params[1],
            warnmin=d.fax64_params[2], warnmax=d.fax64_params[3],
            speed=d.fax64_params[12],
        )
        maxstat = max(maxstat, stat)
    d.v2x64_estatus = maxstat
    d.v2x64_nerrid = d.v2x64_estatus >> 15 if d.v2x64_estatus[[31]] else 0
    d.v2x64_pctl, d.v2x64_pvalue = handle_pctl(
        d.v2x64_pctl, d.v2x64_pvalue, d.fax64_params, d.table64,
    )

    maxstat = 0
    for i in range(16):
        d.v16x64_values[i], d.v16x64_targets[i], stat = advance64(
            d.v16x64_values[i], d.v16x64_targets[i], d.v16x64_estatus,
            usermin=d.fax64_params[0], usermax=d.fax64_params[1],
            warnmin=d.fax64_params[2], warnmax=d.fax64_params[3],
            speed=d.fax64_params[12],
        )
        maxstat = max(maxstat, stat)
    d.v16x64_estatus = maxstat
    d.v16x64_nerrid = d.v16x64_estatus >> 15 if d.v16x64_estatus[[31]] else 0
    d.v16x64_pctl, d.v16x64_pvalue = handle_pctl(
        d.v16x64_pctl, d.v16x64_pvalue, d.fax64_params, d.table64,
    )

    maxstat = 0
    for i in range(2):
        d.v2x32_values[i], d.v2x32_targets[i], stat = advance64(
            d.v2x32_values[i], d.v2x32_targets[i], d.v2x32_estatus,
            usermin=d.fax32_params[0], usermax=d.fax32_params[1],
            warnmin=d.fax32_params[2], warnmax=d.fax32_params[3],
            speed=d.fax32_params[12],
        )
        maxstat = max(maxstat, stat)
    d.v2x32_estatus = maxstat
    d.v2x32_nerrid = d.v2x32_estatus >> 15 if d.v2x32_estatus[[31]] else 0
    d.v2x32_pctl, d.v2x32_pvalue = handle_pctl(
        d.v2x32_pctl, d.v2x32_pvalue, d.fax32_params, d.table32,
    )

    maxstat = 0
    for i in range(16):
        d.v16x32_values[i], d.v16x32_targets[i], stat = advance64(
            d.v16x32_values[i], d.v16x32_targets[i], d.v16x32_estatus,
            usermin=d.fax32_params[0], usermax=d.fax32_params[1],
            warnmin=d.fax32_params[2], warnmax=d.fax32_params[3],
            speed=d.fax32_params[12],
        )
        maxstat = max(maxstat, stat)
    d.v16x32_estatus = maxstat
    d.v16x32_nerrid = d.v16x32_estatus >> 15 if d.v16x32_estatus[[31]] else 0
    d.v16x32_pctl, d.v16x32_pvalue = handle_pctl(
        d.v16x32_pctl, d.v16x32_pvalue, d.fax32_params, d.table32,
    )

    maxstat = 0
    for i in range(2):
        d.lv2x32_values[i], d.lv2x32_targets[i], stat = advance32(
            d.lv2x32_values[i], d.lv2x32_targets[i], d.lv2x32_status,
            usermin=d.fax32_params[0], usermax=d.fax32_params[1],
            warnmin=d.fax32_params[2], warnmax=d.fax32_params[3],
            speed=d.fax32_params[12],
        )
        maxstat = max(maxstat, stat)
    d.lv2x32_status = maxstat
    d.lv2x32_pctl, d.lv2x32_pvalue = handle_pctl(
        d.lv2x32_pctl, d.lv2x32_pvalue, d.fax32_params, d.table32,
    )

    maxstat = 0
    for i in range(16):
        d.lv16x32_values[i], d.lv16x32_targets[i], stat = advance32(
            d.lv16x32_values[i], d.lv16x32_targets[i], d.lv16x32_status,
            usermin=d.fax32_params[0], usermax=d.fax32_params[1],
            warnmin=d.fax32_params[2], warnmax=d.fax32_params[3],
            speed=d.fax32_params[12],
        )
        maxstat = max(maxstat, stat)
    d.lv16x32_status = maxstat
    d.lv16x32_pctl, d.lv16x32_pvalue = handle_pctl(
        d.lv16x32_pctl, d.lv16x32_pvalue, d.fax32_params, d.table32,
    )

    # handle msgio
    def rot13(c):
        if (65 <= c <= 77) or (97 <= c <= 109):
            return c + 13
        if (78 < c <= 90) or (110 <= c <= 122):
            return c - 13
        return c

    if d.msg_mbox >> 13 == 0:
        v.msglen = 0
        d.msg_mbox = 1 << 13
    elif d.msg_mbox >> 13 == 2:
        # partial request -> put to stack
        d.msg_mbox = 3 << 13
        for i in range(MSGIO_NDATA):
            c = d.msg_data[i]
            if c:
                if v.msglen >= 79:
                    # Error: request too long
                    d.msg_mbox = 7 << 13
                    break
                v.msg[v.msglen] = rot13(c)
                v.msglen = v.msglen + 1
            else:
                break
    elif d.msg_mbox >> 13 == 4:
        # final request part-> put to stack
        for i in range(MSGIO_NDATA):
            c = d.msg_data[i]
            if c:
                if v.msglen >= 79:
                    # Error: request too long
                    d.msg_mbox = 7 << 13
                    break
                v.msg[v.msglen] = rot13(c)
                v.msglen = v.msglen + 1
            else:
                break
    if d.msg_mbox >> 13 in (4, 6):
        if v.msglen > 0:
            # transfer size
            ts = min(MSGIO_NDATA, int(v.msglen))
            memcpy(adr(d.msg_data[0]), adr(v.msg[0]), ts)
            v.msglen = v.msglen - ts
            if v.msglen:
                memcpy(adr(v.msg[0]), adr(v.msg[MSGIO_NDATA]), v.msglen)
            d.msg_mbox = (5 << 13 if v.msglen else 1 << 13) + ts
        else:
            d.msg_mbox = 7 << 13

    # handle table lines
    if d.table32_req_row != d.table32_act_row:
        # write back
        memcpy(
            adr(d.table32) + 40 * d.table32_act_row,
            adr(d.table32_line),
            sizeof(d.table32_line),
        )
        # avoid illegal values for req_row
        if not 0 <= d.table32_req_row <= 9:
            d.table32_req_row = d.table32_act_row
        # read from table
        memcpy(
            adr(d.table32_line),
            adr(d.table32) + 40 * d.table32_req_row,
            sizeof(d.table32_line),
        )
        d.table32_act_row = d.table32_req_row

    if d.table64_req_row != d.table64_act_row:
        # write back
        memcpy(
            adr(d.table64) + 80 * d.table64_act_row,
            adr(d.table64_line),
            sizeof(d.table64_line),
        )
        # avoid illegal values for req_row
        if not 0 <= d.table64_req_row <= 9:
            d.table64_req_row = d.table64_act_row
        # read from table
        memcpy(
            adr(d.table64_line),
            adr(d.table64) + 80 * d.table64_req_row,
            sizeof(d.table64_line),
        )
        d.table64_act_row = d.table64_req_row


@program()
def Init(_v):
    i = PLC_DESCRIPTOR.devices
    while i:
        DESCRIPTOR_ARRAY[i].name = typecode_description(DESCRIPTOR_ARRAY[i].typecode)[:DESCRIPTOR_SLOT_SIZE - 21]
        i = DESCRIPTOR_ARRAY[i].prev


@program(is_initialized=Var(boolean, default=False))
def Main(v):
    if not v.is_initialized:
        Init()
        v.is_initialized = True
    Indexer()
    Implementation()



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Implements communication via ADS, the Beckhoff universal protocol."""

import errno
import re
import socket
import time
from struct import Struct, pack

from zapf import ApiError, CommError
from zapf.proto import Protocol

# Allow omitting the final 2 NetID bytes if it's 1.1.
ADS_ADDR_RE = re.compile(r'ads://(.*?)/(\d+\.\d+\.\d+\.\d+(\.\d+\.\d+)?)?:(\d+)$')

# TCP port to use for ADS.
ADS_PORT = 0xBF02

# AMS header including AMS/TCP header.
HEADER = Struct('<HIQQHHIII')
# Reply to device info command.
DEVINFO = Struct('<BBH16s')

HEADER_SIZE = HEADER.size
AMS_HEADER_SIZE = HEADER_SIZE - 6

# ADS commands.
ADS_DEVINFO = 1
ADS_READ    = 2
ADS_WRITE   = 3
ADS_STATUS  = 4

# Formats for payload.
ADS_ADR_LEN = Struct('<III')
ADS_ERRID = Struct('<I')

# Index group for %M memory area.
INDEXGROUP_M = 0x4020

# UDP port
BECKHOFF_UDP_PORT = 0xBF03
# UDP message to set a route.
UDP_ROUTE_MESSAGE = Struct(
    '<I'     # magic
    '4x'     # pad
    'I'      # operation=6
    'Q'      # source netaddr
    'I'      # nitems=5
    'HH25s'  # desig=12 (routename), strlen=25, content
    'HH6s'   # desig=7 (netid), len=6, content
    'HH14s'  # desig=13 (username), len=14, content
    'HH2s'   # desig=2 (password), len=2, content
    'HH16s', # desig=5 (host), len=16, content
)
# UDP message to get system info.
UDP_INFO_MESSAGE = Struct(
    '<I'     # magic
    '4x'     # pad
    'I'      # operation=1
    'Q'      # source netaddr
    '4x',    # nitems=0
)
# UDP magic header number.
UDP_MAGIC = 0x71146603
# UDP packet operations and data designators.
UDP_IDENTIFY = 1
UDP_ADD_ROUTE = 6
UDP_PASSWORD = 2
UDP_HOST = 5
UDP_NETID = 7
UDP_ROUTENAME = 12
UDP_USERNAME = 13


def pack_net_id(netidstr, amsport):
    # convert a.b.c.d.e.f to a single integer
    amsnetid = i = 0
    for (i, x) in enumerate(netidstr.split('.')):
        amsnetid |= (int(x) << (8 * i))
    if i != 5:
        raise ApiError('incomplete NetID; use format a.b.c.d.e.f '
                       'with 6 integers in the range 0-255')

    # pack the whole address into a 64-bit integer
    return amsnetid | (amsport << 48)


class ADSProtocol(Protocol):
    OFFSETS = (0,)

    def __init__(self, url, log):
        adr = ADS_ADDR_RE.match(url)
        if not adr:
            raise ApiError('invalid ADS address, must be '
                           'ads://host[:port]/amsnetid:amsport')
        host = adr.group(1)
        if ':' in host:
            host, port = host.split(':')
            port = int(port)
        else:
            port = ADS_PORT

        self._iphostport = (host, port)
        self._socket = None
        self._tried_route = False

        if adr.group(3):    # full AMS netid specified
            netidstr = adr.group(2)
        elif adr.group(2):  # only first 4 specified
            netidstr = adr.group(2) + '.1.1'
        else:               # nothing specified, query TwinCAT
            netidstr = self._query_netid()
        amsport = int(adr.group(4))

        self._amsnetaddr = pack_net_id(netidstr, amsport)
        self._invoke_id = 1  # should be incremented for every request

        super().__init__(url, log)

    def _retry_connect_with_route(self):
        self.log.warning('connection aborted, trying to set a route...')
        self._tried_route = True
        self._set_route(self._socket.getsockname())
        self.connect()

    def connect(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self._socket.connect(self._iphostport)
            self._socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            self._socket.settimeout(5.0)  # TODO: make configurable?
            self._myamsnetaddr = pack_net_id(
                self._socket.getsockname()[0] + '.1.1', 55189)  # guaranteed random
            reply = self._comm(ADS_DEVINFO, b'', DEVINFO.size)
            v1, v2, v3, name = DEVINFO.unpack(reply)
            name = name.partition(b'\0')[0].decode('latin1')
            hw_name = f'{name} {v1}.{v2}.{v3}'
            self.log.info('connected to %s', hw_name)
        except (OSError, ValueError) as err:
            if getattr(err, 'errno', None) == errno.ECONNRESET and not \
               self._tried_route and self._iphostport[1] == ADS_PORT:
                self._retry_connect_with_route()
                return
            raise CommError(f'could not connect to {self.url}: {err}') from err
        except CommError as err:
            # if we get a closed connection immediately, the route is not set
            # up correctly.  Try to fix that by setting a route via UDP.  If
            # the target TCP port is not default, we are not talking directly
            # to TwinCAT, so don't even try in that case.
            if str(err) == 'no data in read' and not self._tried_route and \
               self._iphostport[1] == ADS_PORT:
                self._retry_connect_with_route()
                return
            raise
        self._connect_callback(True)  # noqa: FBT003
        self.connected = True

    def disconnect(self):
        try:
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()
        except (AttributeError, OSError):
            pass
        self._socket = None
        self._connect_callback(False)  # noqa: FBT003
        self.connected = False

    def read(self, addr, length):
        if not self.connected:
            self.reconnect()
        payload = ADS_ADR_LEN.pack(INDEXGROUP_M, addr, length)
        try:
            return self._comm(ADS_READ, payload, 4 + length)[4:]
        except OSError as err:
            self.log.exception('during read')
            self.disconnect()
            raise CommError(f'IO error during read: {err}') from err

    def write(self, addr, data):
        if not self.connected:
            self.reconnect()
        payload = ADS_ADR_LEN.pack(INDEXGROUP_M, addr, len(data)) + data
        try:
            self._comm(ADS_WRITE, payload, 0)
        except OSError as err:
            self.log.exception('during write')
            self.disconnect()
            raise CommError(f'IO error during write: {err}') from err

    def _comm(self, cmd, payload, exp_len):
        """One ADS request-reply cycle."""
        invoke_id, self._invoke_id = self._invoke_id, self._invoke_id + 1
        msg = HEADER.pack(0, AMS_HEADER_SIZE + len(payload),
                          self._amsnetaddr, self._myamsnetaddr, cmd, 0x4,
                          len(payload), 0, invoke_id)
        self._socket.sendall(msg + payload)
        return self._comm_reply(cmd, invoke_id, exp_len)

    def _comm_reply(self, cmd, invoke_id, exp_len):
        # read the reply header
        rephdr = HEADER.unpack(self._read_exact(HEADER_SIZE))

        # check consistency
        if rephdr[0] != 0:
            # inconsistent packet, close connection
            self.disconnect()
            raise CommError('leading null bytes missing, out of sync?')
        data_len = rephdr[1] - AMS_HEADER_SIZE  # payload incl. result field
        if rephdr[6] != data_len:
            # inconsistent packet, close connection
            self.disconnect()
            raise CommError('wrong length in reply header')

        # always read payload from socket, even if packet will be ignored
        data = self._read_exact(data_len)

        # check if packet is not addressed to us
        if rephdr[2] != self._myamsnetaddr:
            targetport = rephdr[2] >> 48
            # don't warn for Status queries for the system manager
            if not (targetport == 10000 and rephdr[4] == ADS_STATUS):
                self.log.warning(f'got packet for AMS port {targetport}, ignoring')
            # try again with next packet
            return self._comm_reply(cmd, invoke_id, exp_len)

        # check if packet has expected contents
        if rephdr[4] != cmd:
            raise CommError('wrong command in reply header')
        if rephdr[5] != 0x5:
            raise CommError('wrong flags in reply header')
        if rephdr[8] != invoke_id:
            raise CommError('wrong InvokeID on reply packet')
        if rephdr[7]:
            raise CommError('error set in reply header: '
                            f'{self._translate_error(rephdr[7])}')
        if data[:4] != b'\x00\x00\x00\x00':
            result = ADS_ERRID.unpack_from(data)[0]
            raise CommError('error set in reply data: '
                            f'{self._translate_error(result)}')
        if data_len != exp_len + 4:
            raise CommError('received unexpected length in reply '
                            f'(expected {exp_len}, got {data_len})')
        return data[4:]

    def _read_exact(self, length):
        """Blockingly read exactly `length` bytes from socket."""
        data = b''
        while len(data) < length:
            try:
                more = self._socket.recv(length - len(data))
            except TimeoutError:
                raise CommError('timeout while reading bytes') from None
            if not more:
                self.disconnect()
                raise CommError('connection closed by peer')
            data += more
        return data

    def _translate_error(self, errorcode):
        return ADS_ERRORS.get(errorcode, f'Unknown code {errorcode:#06x}')

    def _query_netid(self):
        udpsock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        udpsock.settimeout(1.0)
        message = UDP_INFO_MESSAGE.pack(
            UDP_MAGIC,
            UDP_IDENTIFY,
            pack_net_id('1.1.1.1.1.1', 10000),
        )
        udpsock.sendto(message, (self._iphostport[0], BECKHOFF_UDP_PORT))
        try:
            reply = udpsock.recv(1024)
        except socket.timeout:
            raise CommError('timeout while querying TwinCAT NetID') from None
        return '.'.join(map(str, reply[12:18]))

    def _set_route(self, sockname):
        """Try to set up an ADS route on the target."""
        routename = 'zapf-' + sockname[0]
        udpsock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        mynetid = pack('<Q', self._myamsnetaddr)[:6]
        # The default password is different for different TwinCAT versions.
        # Just try both.
        for password in (b'', b'1'):
            message = UDP_ROUTE_MESSAGE.pack(
                UDP_MAGIC,
                UDP_ADD_ROUTE,
                self._myamsnetaddr,
                5,
                UDP_ROUTENAME, 25, routename.encode(),
                UDP_NETID, 6, mynetid,
                UDP_USERNAME, 14, b'Administrator',
                UDP_PASSWORD, 2, password,
                UDP_HOST, 16, sockname[0].encode(),
            )
            udpsock.sendto(message, (self._iphostport[0], BECKHOFF_UDP_PORT))
        time.sleep(0.5)


# https://infosys.beckhoff.com/english.php?content=../content/1033/tc3_ads_intro_howto/374277003.html&id=2736996179007627436
ADS_ERRORS = {
    0x001: 'Internal error',
    0x002: 'No real-time',
    0x003: 'Allocation locked - memory error',
    0x004: 'Mailbox full - ADS message could not be sent',
    0x005: 'Wrong receive HMSG',
    0x006: 'Target port not found, possibly ADS server not started',
    0x007: 'Target machine not found, possibly missing ADS routes',
    0x008: 'Unknown command ID',
    0x009: 'Invalid task ID',
    0x00A: 'No IO',
    0x00B: 'Unknown AMS command',
    0x00C: 'Win32 error',
    0x00D: 'Port not connected',
    0x00E: 'Invalid AMS length',
    0x00F: 'Invalid AMS NetID',
    0x010: 'Low installation level',
    0x011: 'No debugging available',
    0x012: 'Port disabled - system service not started',
    0x013: 'Port already connected',
    0x014: 'AMS Sync Win32 error',
    0x015: 'AMS Sync timeout',
    0x016: 'AMS Sync error',
    0x017: 'AMS Sync no index map',
    0x018: 'Invalid AMS port',
    0x019: 'No memory',
    0x01A: 'TCP send error',
    0x01B: 'Host unreachable',
    0x01C: 'Invalid AMS fragment',
    0x01D: 'TLS send error - secure ADS connection failed',
    0x01E: 'Access denied - secure ADS access denied',

    0x500: 'Router: no locked memory',
    0x501: 'Router: memory size could not be changed',
    0x502: 'Router: mailbox full',
    0x503: 'Router: debug mailbox full',
    0x504: 'Router: port type is unknown',
    0x505: 'Router is not initialized',
    0x506: 'Router: desired port number is already assigned',
    0x507: 'Router: port not registered',
    0x508: 'Router: maximum number of ports reached',
    0x509: 'Router: port is invalid',
    0x50A: 'Router is not active',
    0x50B: 'Router: mailbox full for fragmented messages',
    0x50C: 'Router: fragment timeout occurred',
    0x50D: 'Router: port removed',

    0x700: 'General device error',
    0x701: 'Service is not supported by server',
    0x702: 'Invalid index group',
    0x703: 'Invalid index offset',
    0x704: 'Reading/writing not permitted',
    0x705: 'Parameter size not correct',
    0x706: 'Invalid parameter value(s)',
    0x707: 'Device is not in a ready state',
    0x708: 'Device is busy',
    0x709: 'Invalid OS context -> use multi-task data access',
    0x70A: 'Out of memory',
    0x70B: 'Invalid parameter value(s)',
    0x70C: 'Not found (files, ...)',
    0x70D: 'Syntax error in command or file',
    0x70E: 'Objects do not match',
    0x70F: 'Object already exists',
    0x710: 'Symbol not found',
    0x711: 'Symbol version invalid -> create a new handle',
    0x712: 'Server is in an invalid state',
    0x713: 'AdsTransMode not supported',
    0x714: 'Notification handle is invalid',
    0x715: 'Notification client not registered',
    0x716: 'No more notification handles',
    0x717: 'Notification size too large',
    0x718: 'Device not initialized',
    0x719: 'Device has a timeout',
    0x71A: 'Query interface failed',
    0x71B: 'Wrong interface required',
    0x71C: 'Class ID is invalid',
    0x71D: 'Object ID is invalid',
    0x71E: 'Request is pending',
    0x71F: 'Request is aborted',
    0x720: 'Signal warning',
    0x721: 'Invalid array index',
    0x722: 'Symbol not active -> release handle and try again',
    0x723: 'Access denied',
    0x724: 'No license found -> activate license',
    0x725: 'License expired',
    0x726: 'License exceeded',
    0x727: 'License invalid',
    0x728: 'Invalid system ID in license',
    0x729: 'License not time limited',
    0x72A: 'License issue time in the future',
    0x72B: 'License time period too long',
    0x72C: 'Exception in device specific code -> check each device',
    0x72D: 'License file read twice',
    0x72E: 'Invalid signature',
    0x72F: 'Invalid public key certificate',
    0x730: 'Public key not known from OEM',
    0x731: 'License not valid for this system ID',
    0x732: 'Demo license prohibited',
    0x733: 'Invalid function ID',
    0x734: 'Outside the valid range',
    0x735: 'Invalid alignment',
    0x736: 'Invalid platform level',
    0x737: 'Context - forward to passive level',
    0x738: 'Content - forward to dispatch level',
    0x739: 'Context - forward to real-time',

    0x740: 'General client error',
    0x741: 'Invalid parameter at service',
    0x742: 'Polling list is empty',
    0x743: 'Var connection already in use',
    0x744: 'Invoke ID in use',
    0x745: 'Timeout elapsed -> check route setting',
    0x746: 'Error in Win32 subsystem',
    0x747: 'Invalid client timeout value',
    0x748: 'ADS port not opened',
    0x749: 'No AMS address',
    0x750: 'Internal error in ADS sync',
    0x751: 'Hash table overflow',
    0x752: 'Key not found in hash',
    0x753: 'No more symbols in cache',
    0x754: 'Invalid response received',
    0x755: 'Sync port is locked',

    0x1000: 'Internal error in real-time system',
    0x1001: 'Timer value not valid',
    0x1002: 'Task pointer has invalid value 0',
    0x1003: 'Stack pointer has invalid value 0',
    0x1004: 'Requested task priority already assigned',
    0x1005: 'No free Task Control Block',
    0x1006: 'No free semaphores',
    0x1007: 'No free space in the queue',
    0x100D: 'External sync interrupt already applied',
    0x100E: 'No external sync interrupt applied',
    0x100F: 'External sync interrupt application failed',
    0x1010: 'Call of service function in wrong context',
    0x1017: 'Intel VT-x not supported',
    0x1018: 'Intel VT-x not enabled in BIOS',
    0x1019: 'Missing function in Intel VT-x',
    0x101A: 'Activation of Intel VT-x failed',
}



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

"""Basic test suite for the utils."""

import pytest

from zapf import SpecError
from zapf.spec import decode_bitfield, v_2015_02, v_2021_09


def test_convert_unitcode():
    # 2015_02 version
    assert v_2015_02.convert_unit(0, -2) == '%'  # 0xfe00
    assert v_2015_02.convert_unit(12, 0) == 'deg'
    assert v_2015_02.convert_unit(16, -3) == 'l/s'
    assert v_2015_02.convert_unit(16, -6) == '10^-6 m^3/s'  # ml/s
    assert v_2015_02.convert_unit(4, -3) == 'mm'
    # 2021_09 version
    assert v_2021_09.decode_unit(0xF000) == '%'
    assert v_2021_09.decode_unit(0x0004) == 'deg'
    assert v_2021_09.decode_unit(0xE888) == 'l/s'
    assert v_2021_09.decode_unit(0xD088) == 'ml/s'
    assert v_2021_09.decode_unit(0xE94F) == 'mT/min'
    assert v_2021_09.decode_unit(0x118C) == 'hH/h'
    assert v_2021_09.decode_unit(0x1801) == '10^3 #'
    assert v_2021_09.decode_unit(0x1806) == 'km'
    assert v_2021_09.decode_unit(0x180e) == 'kOhm'
    assert v_2021_09.decode_unit(0x000e) == 'Ohm'

    assert v_2021_09.decode_unit(0xe808) == 'l'
    assert v_2021_09.decode_unit(0xe948) == 'l/min'
    assert v_2021_09.decode_unit(0xe802) == 'mbar'
    assert v_2021_09.decode_unit(0xe806) == 'mm'
    assert v_2021_09.decode_unit(0xf040) == '%/(#)'
    assert v_2021_09.decode_unit(0x0081) == '#/s'
    assert v_2021_09.decode_unit(0x0141) == '#/min'
    assert v_2021_09.decode_unit(0x0041) == ''  # main per main '#/(#)'

    assert v_2021_09.encode_unit('%') == 0xF000
    assert v_2021_09.encode_unit('deg') == 0x0004
    assert v_2021_09.encode_unit('l/s') == 0xE888
    assert v_2021_09.encode_unit('ml/s') == 0xD088
    assert v_2021_09.encode_unit('mT/min') == 0xE94F
    assert v_2021_09.encode_unit('hH/h') == 0x118C
    assert v_2021_09.encode_unit('10^3 #') == 0x1801
    assert v_2021_09.encode_unit('km') == 0x1806
    assert v_2021_09.encode_unit('Ohm') == 14


def test_decode_bitfield():
    assert decode_bitfield(1) == 1
    with pytest.raises(SpecError):
        decode_bitfield(12, 'a string')
    assert decode_bitfield(12, [(0, 9, '', {})]) == '12'
    assert decode_bitfield(12, [(2, 7, 'N', {})]) == 'N=3'
    assert decode_bitfield(12, [(3, 'A'), (2, 'B'), (0, 'C')]) == 'A, B'
    assert decode_bitfield(12, [(0, 3, 'A', {4: 'X'}), (3, 'B')]) == 'A=X, B'
    assert decode_bitfield(12, [(0, 3, 'A', {4: 'X'}),
                                (3, 1, '', {1: 'B'})]) == 'A=X, B'



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

"""Support for scanning the indexer and creating devices."""

from typing import NamedTuple

from zapf import SpecError, spec
from zapf.device import FMT_LIMITS, TYPECODE_MAP, Device
from zapf.io import PlcIO
from zapf.spec import FMT_TO_BASETYPE, v_2015_02
from zapf.spec.v_2021_09 import fix_unit


class DeviceInfo(NamedTuple):
    name: str
    addr: int
    typecode: int
    info: dict


class Scanner:
    """Scanner for PLC devices.

    :param io_or_proto: A connection URI string, a `.PlcIO` instance or
        a `.Protocol` instance.
    :param log: The `logging.Logger` instance to use for this object and
        all derived objects (e.g. devices).

    **Connection URIs** can have the following forms:

    ``ads://host[:port]/amsnetid:amsport``
        Connection to a Beckhoff PLC using the ADS protocol.  The TCP port is
        48898 by default.  The AMS NetID and AMS port are specific to the PLC.
        Note that an AMS router entry must be set on the PLC in order to
        connect.

        Example: ``ads://192.168.201.2/5.18.77.4.1.1:851``

    ``modbus://host[:port]/slaveno``
        Connection to a host that supports the Modbus/TCP protocol.  The TCP
        port is 502 by default.

        Example: ``modbus://192.168.201.2/0``

    ``tango://dbhost:dbport/tango/device/name``
        Connection to a `Tango <https://tango-controls.org>`_ device which in
        turn connects to the PLC.

        The Tango device interface must conform to the `Profibus
        <https://forge.frm2.tum.de/entangle/defs/entangle-master/profibus/>`_
        Entangle interface specification.

        Example: ``tango://192.168.201.2:10000/box/plc/ads``

    ``sim://filepath``
        "Connection" to a software-simulated PLC.  Zapf starts it in the same
        process when the address is requested.  See the `zapf.simulator`
        package for details.
    """

    IGNORE_BAD_DEVICES = True

    def __init__(self, io_or_proto, log):
        self.log = log
        if isinstance(io_or_proto, PlcIO):
            self.io = io_or_proto
        else:
            self.io = PlcIO(io_or_proto, log)

    def get_plc_data(self):
        """Establish communication with the indexer.

        Return an object containing the metadata.  The following attributes are
        defined on this object:

        ``magicstr``
            The PLC :ref:`magic number <pils:magic>`, converted to a string,
            that identifies the PILS protocol revision.  Example: `'2015_02'`.
        ``indexer_addr``
            The byte address of the indexer data structure, relative to the
            common data area.
        ``indexer_size``
            The size of the indexer in bytes.
        ``num_devices``
            The number of devices present.  This information is provided
            optionally by PILS, and can be 0.  In this case, the number of
            devices is found by finding the first device number where the
            indexer returns empty information.
        ``plc_name``
            The "PLC name" string as reported by the PLC.
        ``plc_version``
            The "PLC version" string as reported by the PLC.
        ``plc_author``
            The "PLC author" string as reported by the PLC.

        Can raise `.SpecError` if the PLC does not conform to a supported PILS
        version, or returns invalid information.
        """
        self.io.indexer.detect_plc()
        return self.io.indexer

    def scan_devices(self):
        """Scan the PLC and yield information about devices.

        For each device found in the PLC, this will yield a namedtuple with the
        following items:

        * ``number``: device number (1 to N)
        * ``name``: device name
        * ``addr``: device byte address
        * ``typecode``: device :ref:`type code <pils:type-codes>`
        * ``info``: additional information: a dictionary, see below

        Additional information contains at least the following keys:

        * ``lowlevel``: bool, if the device is marked "low level"
        * ``unit``: string, the unit of the main device value
        * ``absmin`` and ``absmax``: float, the limits of the main value
        * ``params``: a list of parameter names if the device has params
        * ``funcs``: a list of special function names if the device has such
        * ``aux``: a specification of bitfields/flags containing the interpretation
          of the up to 24 freely definable bits in the device status.
          Strings that are not defined are empty.

        Upon further extension of PILS, additional keys can appear.
        """
        # ensure we can talk to the indexer
        self.io.indexer.detect_plc()

        # check which method we need to scan
        method = getattr(self, '_scan_' + self.io.indexer.magicstr, None)
        if not method:
            raise RuntimeError(f'Magic {self.io.indexer.magicstr} is '
                               'supported, but no scanner method available')

        yield from method()

    def get_device(self, devinfo):
        """Return a device object for the given devinfo.

        The *devinfo* typically comes from scanning.

        This will return a `.Device` object of the correct subclass for the
        device's typecode, or ``None`` if the typecode is not supported.
        """
        devcls = Device.class_for(devinfo.typecode)
        if devcls is None:
            self.log.warning('type code %#x is not supported, skipping',
                             devinfo.typecode)
            return None
        return devcls(devinfo.name, devinfo.addr, devinfo.typecode,
                      devinfo.info, self.io, self.log)

    def get_devices(self):
        """Scan the PLC and yield device objects.

        Short form of calling `get_device` for each device returned by
        `scan_devices`.
        """
        for data in self.scan_devices():
            yield self.get_device(data)

    def _scan_2014_07(self):
        indexer = self.io.indexer
        next_addr = indexer.indexer_addr + indexer.indexer_size

        for devnum in range(1, 256):
            typecode = indexer.query_word(devnum, 0)
            if typecode == 0:
                return
            size = 2 * (typecode & 0xFF)
            addr = next_addr

            name = f'device{devnum}'

            aux = [(i, f'AUX{i}') for i in range(12)]

            try:
                tce = TYPECODE_MAP[typecode]
            except KeyError:
                raise SpecError(f'device {name!r} has unsupported typecode '
                                f'{typecode:#06x}') from None

            basetype, width = FMT_TO_BASETYPE.get(tce.value_fmt, (None, 0))
            access = 'ro' if tce.readonly else 'rw'

            # fixup limits, if needed
            absmin = -spec.FLOAT32_MAX
            absmax = spec.FLOAT32_MAX
            limits = FMT_LIMITS.get(tce.value_fmt)
            if limits:
                absmin = max(absmin, limits[0])
                absmax = min(absmax, limits[1])

            # fixup float typed limits for discrete devices
            if tce.value_fmt in 'hHiIqQ':
                absmin = int(absmin)
                absmax = int(absmax)

            info = {
                'unit': '',
                'absmin': absmin, 'absmax': absmax,
                'params': {}, 'funcs': {},
                'aux': aux, 'tables': {}, 'flags': [],
                'errid': [],
                'description': '',
                'access': access,
                'basetype': basetype,
                'width': width,
                'enum_r': None,
                'enum_w': None,
            }
            self.log.info('found device %s at addr %s with type %#x',
                          name, addr, typecode)
            yield DeviceInfo(name, addr, typecode, info)

            next_addr = addr + 4 * ((size + 3)//4)  # round to next multiple of 4

    def _scan_2015_02(self):
        indexer = self.io.indexer
        next_addr = indexer.indexer_addr + indexer.indexer_size

        for devnum in range(1, 256):
            info = indexer.query_infostruct(devnum)
            typecode, size, addr, unit, flags, absmin, absmax, name = info

            # gone past last device?
            if typecode == 0:
                break

            # if there is no valid data in the infostruct, query individually
            if size + addr + flags == 0:
                size = indexer.query_word(devnum, spec.v_2015_02.INFO_SIZE)
                if not size:
                    size = 2 * (typecode & 0xFF)
                addr = indexer.query_word(devnum, spec.v_2015_02.INFO_ADDR)
                if not addr:
                    addr = next_addr
                unit = indexer.query_unit(devnum, spec.v_2015_02.INFO_UNIT)
                flags = None
                lowlevel = False
                absmin = -spec.FLOAT32_MAX
                absmax = spec.FLOAT32_MAX
            else:
                lowlevel = (flags & 0x80000000) != 0

            # this might be empty even if a valid infostruct was present,
            # e.g. if the full name didn't fit behind the 20 previous bytes
            if not name:
                name = indexer.query_string(devnum, spec.v_2015_02.INFO_NAME)

            # extract info about parameters and special functions
            devclass = typecode >> 13
            parameters = []
            functions = []
            if devclass in (1, 2):
                if devclass == 1:  # FlatDevices: a list of indices
                    plist = indexer.query_bytes(devnum, spec.v_2015_02.INFO_PARAMS)
                    param_ids = [p for p in plist if p]
                else:  # ParamDevices: a bitmap of indices
                    param_ids = indexer.query_bitmap(devnum,
                                                     spec.v_2015_02.INFO_PARAMS)
                param_ids.sort()  # ... which *should* be unnecessary
                for p in param_ids:
                    if spec.v_2015_02.is_function(p):
                        functions.append(p)
                    else:
                        parameters.append(p)

            # extract AUX string labels
            aux_strings = [''] * 24
            if flags is None:  # old way, stop reading the first empty one
                for idx in range(24):
                    aux_string = indexer.query_string(
                        devnum, spec.v_2015_02.INFO_AUX1 + idx,
                    )
                    aux_strings[idx] = aux_string.strip()
                    if not aux_string:
                        break
            else:  # new way, only read the relevant (flagged ones)
                for idx in range(24):
                    if flags & (1 << idx):
                        aux_string = indexer.query_string(
                            devnum, spec.v_2015_02.INFO_AUX1 + idx,
                        )
                        aux_strings[idx] = aux_string.strip()

            # XXX: convert to new style! (see _scan_2021_09)
            flags = ['lowlevel'] if lowlevel else []
            aux = []
            for i, s in enumerate(aux_strings):
                if s:
                    # convert to Flag, bitfields dont exist in 2015_02
                    aux.append((i, s))

            try:
                tce = TYPECODE_MAP[typecode]
            except KeyError:
                raise SpecError(f'device {name!r} has unsupported typecode '
                                f'{typecode:#06x}') from None

            params = {}  # derive from parameters
            funcs = {}  # derive from functions
            if tce.num_params:
                if tce.has_pctrl:
                    # convert params and funcs
                    for p in parameters:
                        par = {'idx': p}
                        par.update(v_2015_02.PARAMETER_Specs[p])
                        if 'unit' in par:
                            par['unit'] = fix_unit(par.get('unit', ''), unit)
                        params[par.pop('name')] = par
                    for p in functions:
                        par = {'idx': p}
                        par.update(v_2015_02.FUNC_Specs[p])
                        if par.get('argument'):
                            arg = {}
                            arg.update(par['argument'])
                            if 'unit' in arg:
                                arg['unit'] = fix_unit(arg.get('unit', ''), unit)
                            par['argument'] = arg
                        funcs[par.pop('name')] = par
                else:
                    # convert flat params
                    for i, p in enumerate(parameters):
                        par = {'idx': i}  # parameter slot id
                        par.update(v_2015_02.PARAMETER_Specs[p])
                        if 'unit' in par:
                            par['unit'] = fix_unit(par.get('unit', ''), unit)
                        params[par.pop('name')] = par

            basetype, width = FMT_TO_BASETYPE.get(tce.value_fmt, (None, 0))
            access = 'ro' if tce.readonly else 'rw'

            # fixup limits, if needed
            limits = FMT_LIMITS.get(tce.value_fmt)
            if limits:
                absmin = max(absmin, limits[0])
                absmax = min(absmax, limits[1])

            # fixup float typed limits for discrete devices
            if tce.value_fmt in 'hHiIqQ':
                absmin = int(absmin)
                absmax = int(absmax)

            info = {
                'unit': unit,
                'absmin': absmin, 'absmax': absmax,
                'params': params, 'funcs': funcs,
                'aux': aux, 'tables': {}, 'flags': flags,
                'errid': [(0, 16, 'errid', {})],
                'description': '',
                'access': access,
                'basetype': basetype,
                'width': width,
                'enum_r': None,
                'enum_w': None,
            }
            self.log.info('found device %s at addr %s with type %#x',
                          name, addr, typecode)
            yield DeviceInfo(name, addr, typecode, info)

            next_addr = addr + size

    def _scan_2021_09(self):
        devid = self.io.indexer.plc_descriptor.last_device
        seen_dev_names = set()

        while devid:
            try:
                dev = self.io.indexer.query_descriptor(devid)
            except SpecError as err:
                self.log.error(err)  # noqa: TRY400
                # try to re-read content without resolving
                dev = self.io.indexer.query_descriptor_uncached(devid)
                devid = dev.prev
                continue

            if dev.name in seen_dev_names:
                self.log.debug(f'skipping device {dev.name!r}: a device with '
                               'that name already exists.')
                devid = dev.prev
                continue
            seen_dev_names.add(dev.name)

            if dev.typecode not in TYPECODE_MAP:
                self.log.error(f'device {dev.name!r} at address {dev.address:#06x} '
                               f'has unsupported typecode {dev.typecode:#06x}')
            try:
                info = dev.get_info()
            except SpecError as err:
                self.log.error(f'device {dev.name!r}: {err}')  # noqa: TRY400
                devid = dev.prev
                if self.IGNORE_BAD_DEVICES:
                    continue
                raise SpecError(f'Bad Device {dev.name!r}: {err}') from err

            self.log.info(f'found device {dev.name!r} at addr '
                          f'{dev.address:#06x} with type {dev.typecode:#06x}')
            yield DeviceInfo(info.pop('name'), info.pop('address'),
                             info.pop('typecode'), info)
            devid = dev.prev



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath('..'))  # noqa: PTH100


# -- Project information -----------------------------------------------------

project = 'Zapf'
author = 'Georg Brandl, Enrico Faulhaber'
copyright = '2021, ' + author  # noqa: A001

master_doc = 'index'
default_role = 'py:obj'
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
]
templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

html_logo = '_static/logo.png'
html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']
html_theme_options = {
    'logo_only': True,
}

autodoc_member_order = 'bysource'

intersphinx_mapping = {
    'pils': ('https://forge.frm2.tum.de/public/doc/plc/master/html', None),
}



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

# pylint: disable=invalid-name, unused-argument, no-value-for-parameter
# ruff: noqa: N802, N815

from itertools import count

from zapf.device import typecode_description
from zapf.simulator.funcs import adr, memcpy, memset, sizeof
from zapf.simulator.runtime import (
    Globals,
    Struct,
    Var,
    array,
    boolean,
    byte,
    dword,
    lreal,
    lword,
    program,
    real,
    string,
    word,
)
from zapf.spec.v_2015_02 import is_function

PILS_MAGIC = 2015.02
INDEXER_DATA_SIZE = 34
INDEXER_SIZE = INDEXER_DATA_SIZE + 2
INDEXER_OFFSET = 64
REAL_MAX = 3.402823e38
MSGIO_NDATA = 22


class ST_Indexer(Struct):  # noqa: N801
    Request = Var(word, 0)
    Data = Var(array(word, 1, INDEXER_DATA_SIZE // 2))


class DevicesLayout(Struct):
    # note: with this mapping a sensible sim isn't possible
    # as we don't know which of the overlayed devs was started...

    # used for ai64, ao64, rv64  (re-using aox_target as aix_value, rv64)
    ax64_value = Var(lreal, 0)             # 0..7
    ax64_target = Var(lreal, 0)            # 8..15
    ax64_estatus = Var(dword, 0x10000000)  # 16..19
    ax64_nerrid = Var(word, 0)             # 20..21
    ax64_reserved = Var(word, 0)           # 22..23
    # used for ai32, ao32, sw32 (re-using aox_target as aix_value)
    ax32_value = Var(real, 0)              # 24..27
    ax32_target = Var(real, 0)             # 28..31
    ax32_estatus = Var(dword, 0x10000000)  # 32..35
    # used for legacy_ai32, _ao32, rv32, sw16  (re-using aox_target as aix_value, rv32)
    lax32_value = Var(real, 0)             # 36..39
    lax32_target = Var(real, 0)            # 40..43
    lax32_status = Var(word, 0x1000)       # 44..45
    # used for di16, do16, kv16, sw16 (re-using dox_target as dix_value)
    dx16_value = Var(word, 0)              # 46..47
    dx16_target = Var(word, 0)             # 48..49
    dx16_status = Var(word, 0x1000)        # 50..51
    # used for di32, do32, kv32, sw32  (re-using dox_target as dix_value)
    dx32_value = Var(dword, 0)             # 52..55
    dx32_target = Var(dword, 0)            # 56..59
    dx32_estatus = Var(dword, 0x10000000)  # 60..63
    # used for sdi64, sdo64, di64, do64, kv64 (re-using dox_target as dix_value)
    dx64_value = Var(lword, 0)             # 64..71
    dx64_target = Var(lword, 0)            # 72..79
    dx64_estatus = Var(dword, 0x10000000)  # 80..83
    dx64_nerrid = Var(word, 0)             # 84..85
    dx64_reserved = Var(word, 0)           # 86..87
    # used for all flatin/out in 64 bit (re-using fo_target as fi_value)
    fax64_value = Var(lreal, 50)           # 88..95
    fax64_target = Var(lreal, 50)          # 96..103
    fax64_estatus = Var(dword, 0x10000000) # 104..107
    fax64_nerrid = Var(word, 0)            # 108..109
    fax64_reserved = Var(word, 0)          # 110..111
    # 112..239
    fax64_params = Var(array(lreal, 0, 15),
                       [0, 100, 10, 90, 5, 500, 0, 10, 2, 0, 0.125, 0, 10, 1, 0, 0])

    # used for all pi/po/vectorin/out in 64 (re-using vo_target(s) as vi_value(s))
    # 240..495
    vx64_valuetargets = Var(array(lreal, 0, 31), [0] * 32)
    vx64_estatus = Var(dword, 0x10000000)  # 496..499
    vx64_nerrid = Var(word, 0)             # 500..501
    vx64_pctl = Var(word, 0)               # 502..503
    vx64_pvalue = Var(lreal, 0)            # 504..511
    # used for all pi/po/vectorin/out in 32 (re-using vo_target(s) as vi_value(s))
    # 512..639
    vx32_valuetargets = Var(array(real, 0, 31), [0] * 32)
    vx32_status = Var(word, 0x1000)        # 640..641
    vx32_pctl = Var(word, 0)               # 642..643
    vx32_pvalue = Var(real, 0)             # 644..647
    # used for all flatin/out in 32 bit (re-using fo_target as fi_value)
    fax32_value = Var(real, 50)            # 648..651
    fax32_target = Var(real, 50)           # 652..655
    fax32_status = Var(word, 0x1000)       # 656..657
    fax32_reserved = Var(word, 0)          # 658..659
    # 660..723
    fax32_params = Var(array(real, 0, 15),
                       [0, 100, 10, 90, 5, 500, 0, 10, 2, 0, 0.125, 0, 10, 1, 0, 0])
    # sai/sao in 32/64 sdi/sdo in 16/32/64
    sao64_value = Var(lreal, 0)            # 724..732
    sao64_target = Var(lreal, 0)           # 732..739
    sdo64_value = Var(lword, 0)            # 740..747
    sdo64_target = Var(lword, 0)           # 748..755
    sao32_value = Var(real, 0)             # 756..759
    sao32_target = Var(real, 0)            # 760..763
    sdo32_value = Var(dword, 0)            # 764..767
    sdo32_target = Var(dword, 0)           # 768..771
    sdo16_value = Var(word, 0)             # 772..773
    sdo16_target = Var(word, 0)            # 774..775
    # 1e04 = discrete output with extended status
    dox_value = Var(word, 0)               # 776..777
    dox_target = Var(word, 0)              # 778..779
    dox_estatus = Var(dword, 0)            # 780..783
    # msgio (0x050c)
    msg_mbox = Var(word, 0)                # 784..785
    msg_data = Var(array(byte, 0, MSGIO_NDATA-1))  # 786..807


# ensure byte addresses above are accurate
assert DevicesLayout.sizeof() == 808


class ST_DeviceInfo(Struct):  # noqa: N801
    TypCode = Var(word, 0)
    Size = Var(word, 0)
    Offset = Var(word, 0)
    Unit = Var(word, 0)
    Flags = Var(dword, 0)
    Params = Var(array(word, 1, 16), [0] * 16)
    Name = Var(string(34))
    Aux = Var(array(string(34), 0, 23))
    AbsMax = Var(real, REAL_MAX)
    AbsMin = Var(real, -REAL_MAX)


AUX8 = [f'AUX bit {i}/8' for i in range(8)]
AUX24 = [f'AUX bit {i}/24' for i in range(24)]
# for devices with flat params
PARS_1 = [0x20, 0]
PARS_8 = [0x2120, 0x2322, 0x2524, 0x3328, 0]
PARS_16 = [0x2120, 0x2322, 0x2524, 0x3328, 0x3534, 0x3938, 0x3D3C, 0x4645, 0]
# for devices with parameter interface
PARAM_BITMAP = [0xE, 0, 0x197F, 0xFFB8, 0x73, 0x0, 0x0, 0x0, 0x4221]
WRITEABLE_PARAMS = []
for p in PARS_16:
    WRITEABLE_PARAMS.append(p & 255)
    WRITEABLE_PARAMS.append(p // 256)


class Global(Globals):
    fMagic = Var(real, PILS_MAGIC, at='%MB0')
    iOffset = Var(word, INDEXER_OFFSET, at='%MB4')

    stIndexer = Var(ST_Indexer, at=f'%MB{INDEXER_OFFSET}')

    data = Var(DevicesLayout, at='%MB200')

    # note: Name fields will be overriden later in Init().
    # here they are only a programming/ers hint.
    Devices = Var(array(ST_DeviceInfo, 1, 59), [
        ST_DeviceInfo(TypCode=0x1201, Name='sdi16', Offset=200+772),
        ST_DeviceInfo(TypCode=0x1202, Name='sdi32', Offset=200+764),
        ST_DeviceInfo(TypCode=0x1204, Name='sdi64', Offset=200+740),
        ST_DeviceInfo(TypCode=0x1302, Name='sai32', Offset=200+756),
        ST_DeviceInfo(TypCode=0x1304, Name='sai64', Offset=200+724),
        ST_DeviceInfo(TypCode=0x1401, Name='kw16', Offset=200+48),
        ST_DeviceInfo(TypCode=0x1402, Name='kw32', Offset=200+56),
        ST_DeviceInfo(TypCode=0x1404, Name='kw64', Offset=200+72),
        ST_DeviceInfo(TypCode=0x1502, Name='rv32', Offset=200+40),
        ST_DeviceInfo(TypCode=0x1504, Name='rv64', Offset=200+8),
        ST_DeviceInfo(TypCode=0x1602, Name='sdo16', Offset=200+772),
        ST_DeviceInfo(TypCode=0x1604, Name='sdo32', Offset=200+764),
        ST_DeviceInfo(TypCode=0x1608, Name='sdo64', Offset=200+740),
        ST_DeviceInfo(TypCode=0x1704, Name='sao32', Offset=200+756),
        ST_DeviceInfo(TypCode=0x1708, Name='sao64', Offset=200+724),
        ST_DeviceInfo(TypCode=0x1801, Name='sw16', Offset=200+44, Aux=AUX8),
        ST_DeviceInfo(TypCode=0x1802, Name='sw32', Offset=200+32, Aux=AUX24),
        ST_DeviceInfo(TypCode=0x1a02, Name='di16', Offset=200+48, Aux=AUX8),
        ST_DeviceInfo(TypCode=0x1a04, Name='di32', Offset=200+56, Aux=AUX24),
        ST_DeviceInfo(TypCode=0x1a08, Name='di64', Offset=200+72, Aux=AUX24),
        ST_DeviceInfo(TypCode=0x1b03, Name='lai32', Offset=200+40, Aux=AUX8),
        ST_DeviceInfo(TypCode=0x1b04, Name='ai32', Offset=200+28, Aux=AUX24),
        ST_DeviceInfo(TypCode=0x1b08, Name='ai64', Offset=200+8, Aux=AUX24),
        ST_DeviceInfo(TypCode=0x1e03, Name='do16', Offset=200+46, Aux=AUX8),
        ST_DeviceInfo(TypCode=0x1e06, Name='do32', Offset=200+52, Aux=AUX24),
        ST_DeviceInfo(TypCode=0x1e0c, Name='do64', Offset=200+64, Aux=AUX24),
        ST_DeviceInfo(TypCode=0x1f05, Name='lao32', Offset=200+36, Aux=AUX8),
        ST_DeviceInfo(TypCode=0x1f06, Name='ao32', Offset=200+24, Aux=AUX24),
        ST_DeviceInfo(TypCode=0x1f0c, Name='ao64', Offset=200+0, Aux=AUX24),

        ST_DeviceInfo(TypCode=0x2006, Name='fa32i1', Offset=200+652,
                      Unit=0xfd04, Aux=AUX8, Params=PARS_1),
        ST_DeviceInfo(TypCode=0x2714, Name='fa32i8', Offset=200+652,
                      Unit=0xfd04, Aux=AUX8, Params=PARS_8),
        ST_DeviceInfo(TypCode=0x2f24, Name='fa32i16', Offset=200+652,
                      Unit=0xfd04, Aux=AUX8, Params=PARS_16),

        ST_DeviceInfo(TypCode=0x200c, Name='fa64i1', Offset=200+96,
                      Unit=0xfd04, Aux=AUX24, Params=PARS_1),
        ST_DeviceInfo(TypCode=0x2728, Name='fa64i8', Offset=200+96,
                      Unit=0xfd04, Aux=AUX24, Params=PARS_8),
        ST_DeviceInfo(TypCode=0x2f48, Name='fa64i16', Offset=200+96,
                      Unit=0xfd04, Aux=AUX24, Params=PARS_16),

        ST_DeviceInfo(TypCode=0x3008, Name='fa32o1', Offset=200+648,
                      Unit=0xfd04, Aux=AUX8, Params=PARS_1),
        ST_DeviceInfo(TypCode=0x3716, Name='fa32o8', Offset=200+648,
                      Unit=0xfd04, Aux=AUX8, Params=PARS_8),
        ST_DeviceInfo(TypCode=0x3f26, Name='fa32o16', Offset=200+648,
                      Unit=0xfd04, Aux=AUX8, Params=PARS_16),

        ST_DeviceInfo(TypCode=0x3010, Name='fa64o1', Offset=200+88,
                      Unit=0xfd04, Aux=AUX24, Params=PARS_1),
        ST_DeviceInfo(TypCode=0x372c, Name='fa64o8', Offset=200+88,
                      Unit=0xfd04, Aux=AUX24, Params=PARS_8),
        ST_DeviceInfo(TypCode=0x3f4c, Name='fa64o16', Offset=200+88,
                      Unit=0xfd04, Aux=AUX24, Params=PARS_16),

        ST_DeviceInfo(TypCode=0x4006, Name='pi32', Offset=200+636,
                      Unit=0xfe00, Aux=AUX8, Params=PARAM_BITMAP),
        ST_DeviceInfo(TypCode=0x5008, Name='po32', Offset=200+632,
                      Unit=0xfe00, Aux=AUX8, Params=PARAM_BITMAP),

        ST_DeviceInfo(TypCode=0x400c, Name='pi64', Offset=200+488,
                      Unit=0xfe00, Aux=AUX24, Params=PARAM_BITMAP),
        ST_DeviceInfo(TypCode=0x5010, Name='po64', Offset=200+480,
                      Unit=0xfe00, Aux=AUX24, Params=PARAM_BITMAP),

        ST_DeviceInfo(TypCode=0x4108, Name='v2i32', Offset=200+632,
                      Unit=0xfe00, Aux=AUX8, Params=PARAM_BITMAP),
        ST_DeviceInfo(TypCode=0x4714, Name='v8i32', Offset=200+608,
                      Unit=0xfe00, Aux=AUX8, Params=PARAM_BITMAP),
        ST_DeviceInfo(TypCode=0x4f24, Name='v16i32', Offset=200+576,
                      Unit=0xfe00, Aux=AUX8, Params=PARAM_BITMAP),

        ST_DeviceInfo(TypCode=0x510c, Name='v2o32', Offset=200+624,
                      Unit=0xfe00, Aux=AUX8, Params=PARAM_BITMAP),
        ST_DeviceInfo(TypCode=0x5724, Name='v8o32', Offset=200+576,
                      Unit=0xfe00, Aux=AUX8, Params=PARAM_BITMAP),
        ST_DeviceInfo(TypCode=0x5f44, Name='v16o32', Offset=200+512,
                      Unit=0xfe00, Aux=AUX8, Params=PARAM_BITMAP),

        ST_DeviceInfo(TypCode=0x4110, Name='v2i64', Offset=200+480,
                      Unit=0xfe00, Aux=AUX24, Params=PARAM_BITMAP),
        ST_DeviceInfo(TypCode=0x4728, Name='v8i64', Offset=200+432,
                      Unit=0xfe00, Aux=AUX24, Params=PARAM_BITMAP),
        ST_DeviceInfo(TypCode=0x4f48, Name='v16i64', Offset=200+368,
                      Unit=0xfe00, Aux=AUX24, Params=PARAM_BITMAP),

        ST_DeviceInfo(TypCode=0x5118, Name='v2o64', Offset=200+464,
                      Unit=0xfe00, Aux=AUX24, Params=PARAM_BITMAP),
        ST_DeviceInfo(TypCode=0x5748, Name='v8o64', Offset=200+368,
                      Unit=0xfe00, Aux=AUX24, Params=PARAM_BITMAP),
        ST_DeviceInfo(TypCode=0x5f88, Name='v16o64', Offset=200+240,
                      Unit=0xfe00, Aux=AUX24, Params=PARAM_BITMAP),

        ST_DeviceInfo(TypCode=0x1e04, Name='dox', Offset=200+776),

        ST_DeviceInfo(TypCode=0x0501 + MSGIO_NDATA//2,
                      Name='msg22', Offset=200 + 784),
    ])

    sPLCName = Var(string(34), 'lazy test plc')
    sPLCVersion = Var(string(34), '0.0.1-alpha')
    sPLCAuthor1 = Var(string(34), 'anonymous')
    sPLCAuthor2 = Var(string(34), 'author')
    iCycle = Var(word, 0)


g = Global()


@program(
    nDevices = Var(word),
    devnum = Var(word),
    infotype = Var(word),
    is_initialized = Var(boolean, default=False),
    itemp = Var(byte),
    tempofs = Var(word),
)
def Indexer(v):
    if not v.is_initialized:
        v.tempofs = g.iOffset + sizeof(g.stIndexer)
        v.nDevices = sizeof(g.Devices) // sizeof(g.Devices[1])
        v.itemp = 1
        while v.itemp <= v.nDevices:
            dev = g.Devices[v.itemp]
            for i in range(24):
                dev.Flags[[i]] = len(dev.Aux[i]) > 0
            dev.Size = max(dev.Size, (dev.TypCode & 0xFF) << 1)
            if dev.Offset == 0:
                dev.Offset = v.tempofs
            else:
                v.tempofs = dev.Offset
            v.tempofs += dev.Size
            v.itemp += 1
        v.is_initialized = True

    if g.fMagic != PILS_MAGIC:
        g.fMagic = PILS_MAGIC
    if g.iOffset != INDEXER_OFFSET:
        g.iOffset = INDEXER_OFFSET

    req_num = g.stIndexer.Request & 0x7FFF
    v.devnum = req_num & 0xFF
    v.infotype = req_num >> 8

    data = g.stIndexer.Data

    # short cut, if there is no request, do nothing.
    if g.stIndexer.Request[[15]] == 1:
        if v.infotype == 127:
            data[1] = g.iCycle

        g.iCycle += 1
        return

    memset(adr(data), 0, sizeof(data))

    if v.devnum == 0:
        if v.infotype == 0:
            data[1] = 0
            data[2] = sizeof(g.stIndexer)
            data[3] = g.iOffset
            data[4] = 0
            data[5] = v.nDevices
            data[6] = 0x8200
        elif v.infotype == 1:
            data[1] = sizeof(g.stIndexer)
        elif v.infotype == 4:
            memcpy(adr(data), adr(g.sPLCName), min(sizeof(g.sPLCName), sizeof(data)))
        elif v.infotype == 5:
            memcpy(
                adr(data), adr(g.sPLCVersion), min(sizeof(g.sPLCVersion), sizeof(data)),
            )
        elif v.infotype == 6:
            memcpy(
                adr(data), adr(g.sPLCAuthor1), min(sizeof(g.sPLCAuthor1), sizeof(data)),
            )
        elif v.infotype == 7:
            memcpy(
                adr(data), adr(g.sPLCAuthor2), min(sizeof(g.sPLCAuthor2), sizeof(data)),
            )
    elif v.devnum <= v.nDevices:
        dev = g.Devices[v.devnum]
        if v.infotype == 0:
            data[1] = dev.TypCode
            data[2] = dev.Size
            data[3] = dev.Offset
            data[4] = dev.Unit
            data[5] = dev.Flags
            data[6] = dev.Flags >> 16
            memcpy(adr(data[7]), adr(dev.AbsMin), sizeof(dev.AbsMin))
            memcpy(adr(data[9]), adr(dev.AbsMax), sizeof(dev.AbsMax))
            memcpy(
                adr(data[11]), adr(dev.Name), min(sizeof(dev.Name), sizeof(data) - 20),
            )
        elif v.infotype == 1:
            data[1] = dev.Size
        elif v.infotype == 2:
            data[1] = dev.Offset
        elif v.infotype == 3:
            data[1] = dev.Unit
        elif v.infotype == 4:
            memcpy(adr(data), adr(dev.Name), min(sizeof(dev.Name), sizeof(data)))
        elif v.infotype == 8:
            memcpy(adr(data), adr(dev.descr), sizeof(dev.descr))
        elif v.infotype == 15:
            memcpy(adr(data), adr(dev.Params), min(sizeof(dev.Params), sizeof(data)))
        elif v.infotype >= 0x10 and v.infotype <= 0x27:
            memcpy(
                adr(data),
                adr(dev.Aux[v.infotype - 0x10]),
                min(sizeof(dev.Aux[v.infotype - 0x10]), sizeof(data)),
            )

    if v.infotype == 127:
        data[1] = g.iCycle

    g.stIndexer.Request[[15]] = 1
    g.iCycle += 1


# helper
def advance32(value, target, status, usermin=0.0, usermax=100.0,
              warnmin=10.0, warnmax=90.0, speed=10.0):
    state = status >> 12
    reason = 0
    if state == 0:
        state = 1, 0
    elif state in [1, 3, 7]:
        state = 1 if warnmin <= value <= warnmax else 3
    elif state == 5:
        state, reason = (6, 0) if usermin <= target <= usermax else (8, 1)
    elif state == 6:
        if value == target:
            state, reason = 1, 0
        value += max(min((target - value), speed / 20), -speed / 20)
    target = max(min(usermax, target), usermin)
    if value < warnmin:
        reason |= 4
    if value > warnmax:
        reason |= 8

    return value, target, (state << 12) | (reason << 8)


# same as above but for 32 bit status fields
def advance64(value, target, status, *args, **kwds):
    v, t, s = advance32(value, target, status >> 16, *args, **kwds)
    return v, t, s << 16


# pylint: disable=too-many-return-statements
def handle_pctl(pctl, pvalue, valuestore):
    if pctl & 0x8000:
        # nothing to do
        return pctl, pvalue
    pnum = pctl & 8191
    if pnum > len(PARAM_BITMAP) * 16:
        # non exisiting param -> ERR_NO_IDX
        return 0xA000 | pnum, pvalue
    if (PARAM_BITMAP[pnum // 16] >> (pnum & 15)) & 1 == 0:
        # Param not in our declared param list -> ERR_NO_IDX
        return 0xA000 | pnum, pvalue
    # param exists, can at least be read
    cmd = pctl >> 13
    if cmd == 0:  # INIT
        return 0x8000, 0
    if cmd == 1:  # DO_READ
        if pnum in WRITEABLE_PARAMS:
            # return value and DONE
            return 0x8000 | pnum, valuestore[WRITEABLE_PARAMS.index(pnum)]
        # invent value and DONE
        # HACK: pnums <= 30 should be ints: fake this by returning 0
        return 0x8000 | pnum, 0 if pnum <= 30 else pnum ** 1.5
    if cmd == 2:  # DO_WRITE
        if pnum in WRITEABLE_PARAMS:
            valuestore[WRITEABLE_PARAMS.index(pnum)] = pvalue
            return 0x6000 | pnum, pvalue  # -> BUSY
        if is_function(pnum):
            # go to BUSY for some and ERR_RETRY for others
            if pnum in (128, 133):
                return 0x6000 | pnum, pvalue
            return 0xE000 | pnum, pvalue  # ERR_RETRY
        # no function, no storage -> ERR_READ_ONLY
        return 0xC000 | pnum, pvalue
    if cmd == 3:  # BUSY
        return 0x2000 | pnum, pvalue
    raise RuntimeError(
        f'handle_pctl({pctl}, {pvalue}, {valuestore}): This should never happen!',
    )


def handle_sxo(value, target, step):
    """Handle the simple a/d outputs."""
    if value < target:
        return min(value + step, target)
    if value > target:
        return max(value - step, target)
    return value


def handle_dx16(value, target, status):
    if status >> 12 == 5:
        return target, 0x6000
    return value, 0x1000


def handle_dx32(value, target, status):
    v, s = handle_dx16(value, target, status >> 16)
    return v, s << 16


@program(
    msg=Var(array(byte, 0, 80)),
    msglen=Var(byte, 0),
)
def Implementation(v):
    d = g.data

    d.ax64_value, d.ax64_target, d.ax64_estatus = advance64(
        d.ax64_value, d.ax64_target, d.ax64_estatus,
    )
    d.ax64_nerrid = d.ax64.estatus >> 15 if d.ax64_estatus[[31]] else 0

    d.ax32_value, d.ax32_target, d.ax32_estatus = advance64(
        d.ax32_value, d.ax32_target, d.ax32_estatus,
    )

    d.lax32_value, d.lax32_target, d.lax32_status = advance32(
        d.lax32_value, d.lax32_target, d.lax32_status,
    )

    d.dx16_target &= 0xFF
    d.dx16_value, d.dx16_status = handle_dx16(
        d.dx16_value, d.dx16_target, d.dx16_status,
    )

    d.dx32_target &= 0xFFFF
    d.dx32_value, d.dx32_estatus = handle_dx32(
        d.dx32_value, d.dx32_target, d.dx32_estatus,
    )

    d.dx64_target &= 0xFFFFFFFF
    d.dx64_value, d.dx64_estatus = handle_dx32(
        d.dx64_value, d.dx64_target, d.dx64_estatus,
    )
    d.dx64_nerrid = d.dx16_value

    d.dox_value, d.dox_estatus = handle_dx32(d.dox_value, d.dox_target, d.dox_estatus)

    for i in range(16):
        d.fax64_params[i] = max(0, d.fax64_params[i])
    d.fax64_value, d.fax64_target, d.fax64_estatus = advance64(
        d.fax64_value, d.fax64_target, d.fax64_estatus,
        usermin=d.fax64_params[0], usermax=d.fax64_params[1],
        warnmin=d.fax64_params[2], warnmax=d.fax64_params[3],
        speed=d.fax64_params[12],
    )
    d.fax64_nerrid = d.fax64.estatus >> 15 if d.fax64_estatus[[31]] else 0

    for i in range(16):
        d.fax32_params[i] = max(0, d.fax32_params[i])
    d.fax32_value, d.fax32_target, d.fax32_status = advance32(
        d.fax32_value, d.fax32_target, d.fax32_status,
        usermin=d.fax32_params[0], usermax=d.fax32_params[1],
        warnmin=d.fax32_params[2], warnmax=d.fax32_params[3],
        speed=d.fax32_params[12],
    )

    d.vx32_valuetargets[30], d.vx32_valuetargets[31], d.vx32_status = advance32(
        d.vx32_valuetargets[30], d.vx32_valuetargets[31], d.vx32_status,
    )
    if d.vx32_status[[14]]:  # while start/busy/stop
        for i in range(30):
            d.vx32_valuetargets[i] = d.vx32_valuetargets[30]
    # handle pctl.
    d.vx32_pctl, d.vx32_pvalue = handle_pctl(d.vx32_pctl, d.vx32_pvalue, d.fax32_params)

    d.vx64_valuetargets[30], d.vx64_valuetargets[31], d.vx64_estatus = advance64(
        d.vx64_valuetargets[30], d.vx64_valuetargets[31], d.vx64_estatus,
    )
    if d.vx64_estatus[[30]]:  # while start/busy/stop
        for i in range(30):
            d.vx64_valuetargets[i] = d.vx64_valuetargets[30]
    d.vx64_nerrid = d.vx64.estatus >> 15 if d.vx64_estatus[[31]] else 0
    # handle pctl.
    d.vx64_pctl, d.vx64_pvalue = handle_pctl(d.vx64_pctl, d.vx64_pvalue, d.fax64_params)

    d.sao64_value = handle_sxo(d.sao64_value, d.sao64_target, 0.12)
    d.sao32_value = handle_sxo(d.sao32_value, d.sao32_target, 0.12)

    d.sdo64_value = handle_sxo(d.sdo64_value, d.sdo64_target, 1)
    d.sdo32_value = handle_sxo(d.sdo32_value, d.sdo32_target, 1)
    d.sdo16_value = handle_sxo(d.sdo16_value, d.sdo16_target, 1)

    # handle msgio
    def rot13(c):
        if (65 <= c <= 77) or (97 <= c <= 109):
            return c + 13
        if (78 < c <= 90) or (110 <= c <= 122):
            return c - 13
        return c

    if d.msg_mbox >> 13 == 0:
        v.msglen = 0
        d.msg_mbox = 1 << 13
    elif d.msg_mbox >> 13 == 2:
        # partial request -> put to stack
        d.msg_mbox = 3 << 13
        for i in range(MSGIO_NDATA):
            c = d.msg_data[i]
            if c:
                if v.msglen >= 79:
                    # Error: request too long
                    d.msg_mbox = 7 << 13
                    break
                v.msg[v.msglen] = rot13(c)
                v.msglen = v.msglen + 1
            else:
                break
    elif d.msg_mbox >> 13 == 4:
        # final request part-> put to stack
        for i in range(MSGIO_NDATA):
            c = d.msg_data[i]
            if c:
                if v.msglen >= 79:
                    # Error: request too long
                    d.msg_mbox = 7 << 13
                    break
                v.msg[v.msglen] = rot13(c)
                v.msglen = v.msglen + 1
            else:
                break
    if d.msg_mbox >> 13 in (4, 6):
        if v.msglen > 0:
            # transfer size
            ts = int(min(MSGIO_NDATA, v.msglen))
            memcpy(adr(d.msg_data[0]), adr(v.msg[0]), ts)
            v.msglen = v.msglen - ts
            if v.msglen:
                memcpy(adr(v.msg[0]), adr(v.msg[MSGIO_NDATA]), v.msglen)
            d.msg_mbox = (5 << 13 if v.msglen else 1 << 13) + ts
        else:
            d.msg_mbox = 7 << 13


@program()
def Init(_v):
    for i in count(1):
        try:
            g.Devices[i].Name = typecode_description(g.Devices[i].TypCode)
        except RuntimeError:  # noqa: PERF203
            break


@program(
    is_initialized = Var(boolean, default=False),
)
def Main(v):
    if not v.is_initialized:
        Init()
        v.is_initialized = True
    Indexer()
    Implementation()



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

"""Object for handling table data."""

import time

from zapf import ApiError, CommError, SpecError
from zapf.spec import ParamCMDs, v_2021_09
from zapf.spec.v_2021_09 import get_valueinfo


class Table:
    def __init__(self, io, dev, name, info):
        self.io = io
        self.dev = dev
        self.name = name
        self.info = info
        self._rows = info['last_row'] + 1
        self._cols = info['last_column'] + 1
        self._addr = info['idx']
        self._colinfo = {}
        self._last_row = None
        for col, column in enumerate(info['header']):
            self._colinfo[col] = self._colinfo[column['name']] = \
                (col, get_valueinfo(column, self.dev.value_size))

        if info['accesstype'] == v_2021_09.TABLE_ACCESS_PCTL:
            if not dev.pctrl_addr:
                raise SpecError('device has no paramctrl, yet a table '
                                f'{name!r} with that accesstype is declared')
        elif dev.value_size < 4:
            raise SpecError('tables are not allowed on 16-bit devices')
        if info['accesstype'] == v_2021_09.TABLE_ACCESS_MEM_1D:
            io.register_cache_range(self._addr, 4 + self._rows*self.dev.value_size)
        elif info['accesstype'] == v_2021_09.TABLE_ACCESS_MEM_2D:
            io.register_cache_range(self._addr,
                                    self._rows * self._cols * self.dev.value_size)

    def get_size(self):
        """Return the table size as (rows, cols)."""
        return (self._rows, self._cols)

    def list_columns(self):
        """Read the table's columns names (in order)."""
        return [col['name'] for col in self.info['header']]

    def get_column_valueinfo(self, cname):
        return self._colinfo[cname][1]

    def get_cell_raw(self, row, cname):
        """Get the value of a given cell.

        ``cname`` is either the name of a column or a numeric index into the
        header list obtained from `.list_columns`.

        The type of the value depends on the column, it can be int or float.
        enums are not translated.
        """
        if row >= self._rows:
            raise ApiError(f"can't access row {row}, it must be < {self._rows}")
        if row < 0:
            raise ApiError(f"can't access row {row}, it must be >= 0")
        try:
            col, value_info = self._colinfo[cname]
        except KeyError:
            if isinstance(cname, str):
                raise ApiError(f'column {cname!r} does not exist') from None
            raise ApiError(f"can't access column {cname}, it must be "
                           f'< {self._cols}') from None
        fmt = value_info.fmt
        accesstype = self.info['accesstype']
        cell_number = self._cols * row + col
        if accesstype == v_2021_09.TABLE_ACCESS_PCTL:
            # pylint: disable=protected-access
            pctrl_state, pval = self.dev._get_param(self._addr + cell_number, fmt)  # noqa: SLF001
            if pctrl_state is ParamCMDs.DONE:
                return pval
            raise CommError(f'getting table {self.name!r} cell @ '
                            f'row={row} column={cname!r} '
                            f'failed with ParamCMD {ParamCMDs[pctrl_state]!r}')
        if accesstype == v_2021_09.TABLE_ACCESS_MEM_1D:
            # select right row
            act_row_addr = self._addr
            req_row_addr = self._addr + 2
            addr = self._addr + 4 + self.dev.value_size * col
            tries = 10
            while self.io.read_u16(act_row_addr) != row:
                self.io.write_u16(req_row_addr, row)
                tries -= 1
                if not tries:
                    raise CommError(f'mailbox for table {self.name!r} not responding')
                time.sleep(0.001)
            return self.io.read_fmt(addr, fmt)[0]
        if accesstype == v_2021_09.TABLE_ACCESS_MEM_2D:
            addr = self._addr + self.dev.value_size * cell_number
            return self.io.read_fmt(addr, fmt)[0]
        raise SpecError(f'unknown accesstype {accesstype!r}')

    def get_cell(self, row, cname):
        """Get the value of a given cell.

        Like `get_cell_raw`, but enums are translated.
        """
        value = self.get_cell_raw(row, cname)
        _, value_info = self._colinfo[cname]
        if value_info.enum_r:
            value = str(value_info.enum_r.get(value, value))
        return value

    def set_cell_raw(self, row, cname, value):
        """Set the given cell to a new value.

        Returns the read-back value of that cell.  The type of the value
        depends on the column name, it can be int or float, enums are not
        translated.

        If the table has no column with index or name ``cname``,
        `.ApiError` is raised.
        If the cell is readonly, `.ApiError` is raised.
        """
        if row >= self._rows:
            raise ApiError(f"can't access row {row}, it must be < {self._rows}")
        if row < 0:
            raise ApiError(f"can't access row {row}, it must be >= 0")
        try:
            col, value_info = self._colinfo[cname]
        except KeyError:
            if isinstance(cname, str):
                raise ApiError(f'column {cname!r} does not exist') from None
            raise ApiError(f"can't access column {cname}, it must be < "
                           f'{self._cols}') from None
        if value_info.readonly:
            raise ApiError(f'table {self.name!r} column {cname!r} is read-only')
        fmt = value_info.fmt
        accesstype = self.info['accesstype']
        cell_number = self._cols * row + col
        self._last_row = row
        if accesstype == v_2021_09.TABLE_ACCESS_PCTL:
            # pylint: disable=protected-access
            pctrl_state, pval = self.dev._set_param(  # noqa: SLF001
                self._addr + cell_number, fmt, value)
            if pctrl_state is ParamCMDs.DONE:
                return pval
            raise CommError(f'setting table {self.name!r} cell @ '
                            f'row={row} column={cname!r} '
                            f'failed with ParamCMD {ParamCMDs[pctrl_state]!r}')
        if accesstype == v_2021_09.TABLE_ACCESS_MEM_1D:
            # select right row
            act_row_addr = self._addr
            req_row_addr = self._addr + 2
            addr = self._addr + 4 + self.dev.value_size * col

            for i in range(10):
                if self.io.read_u16(act_row_addr) == row:
                    break
                self.io.write_u16(req_row_addr, row)
                time.sleep(0.001 * i)
            else:
                raise CommError(f'mailbox for table {self.name!r} not responding')
            self.io.write_fmt(addr, fmt, value)
            return self.io.read_fmt(addr, fmt)[0]
        if accesstype == v_2021_09.TABLE_ACCESS_MEM_2D:
            addr = self._addr + self.dev.value_size * cell_number
            self.io.write_fmt(addr, fmt, value)
            return self.io.read_fmt(addr, fmt)[0]
        raise SpecError(f'unknown accesstype {accesstype!r}')

    def set_cell(self, row, cname, value):
        """Set the given cell to a new value.

        Like `set_cell_raw`, but enums are translated and limits are checked.
        """
        try:
            col, value_info = self._colinfo[cname]
        except KeyError:
            if isinstance(cname, str):
                raise ApiError(f'column {cname!r} does not exist') from None
            raise ApiError(f"can't access column {cname}, it must be "
                           f'< {self._cols}') from None
        if isinstance(value, str):
            if value_info.enum_w and value not in value_info.enum_w:
                raise ApiError(f'value {value!r} is not one of the allowed '
                               f'enums (try one of {list(value_info.enum_w)})')
            if not value_info.enum_w:
                raise ApiError(f'string values {value!r} not allowed for cell')
            value = value_info.enum_w[value]
        elif (value < value_info.min_value or
              value > value_info.max_value):
            raise ApiError(f'{value} is not within cell limits '
                           f'{value_info.min_value}..'
                           f'{value_info.max_value}')
        self.set_cell_raw(row, col, value)

    # support >>>with table_obj as t: t.set_cell(...) syntax
    def __enter__(self):
        self._last_row = None
        return self

    def __exit__(self, _type, _value, _traceback):
        if self._last_row is not None and self._rows > 1:
            row = self._last_row
            # force unbuffer of last write by reading from another row
            self.get_cell_raw(row-1 if row else 1, 0)
            self._last_row = None
        return False



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

"""Some helper types and functions."""

import itertools


class EmptyBitField:
    """Integer-like with support for accessing and setting bit ranges.

    This is a baseclass. To use, create a nonempty subclass with BitFields().
    """

    def __init__(self, value=0):
        self._value = int(value)

    def __call__(self, value):
        self._value = int(value)

    def __int__(self):
        return int(self._value)

    def __repr__(self):
        return self.__class__.__name__ + '(' + \
            ', '.join(f'{name}={getattr(self, name)}'
                      for name in self.__fields__) + ')'


def BitFields(**fields):  # noqa: N802
    """Let one access individual bit ranges in a single integer.

    The fields given as keyword arguments are either single bit number or an
    inclusive range of bits given as a (start, end) tuple.  Ranges may overlap.

    Usage example:

    >>> BF = BitFields(LSB=0, x=(3, 0), y=(7, 0), MSB=7)
    >>> bf = BF()
    >>> bf(35)         # update internal state to new value
    >>> bf.LSB         # check LSB (bit0)
    1
    >>> bf.x = 2       # modify bits 3..0 to be 2 = 0b0010
    >>> bf.y           # check bits 0..7 = 0x22 = 34
    34
    >>> int(bf)        # print internal state
    34
    """
    items = {}
    for name, spec in fields.items():
        if isinstance(spec, tuple):
            mask = (1 << (max(spec) - min(spec) + 1)) - 1
            shift = min(spec)
        else:
            mask = 1
            shift = spec

        def getbits(self, shift=shift, mask=mask):
            # pylint: disable=protected-access
            return (self._value >> shift) & mask

        def setbits(self, value, shift=shift, mask=mask):
            # pylint: disable=protected-access
            masked = (self._value >> shift) & mask
            self._value ^= (masked ^ (value & mask)) << shift

        items[name] = property(getbits, setbits, doc=name)
    items['__fields__'] = fields
    return type('BitFields', (EmptyBitField,), items)


class UncasedMap(dict):
    """A dict subclass that allows two-way lookup of unique key-value pairs.

    In addition to the initial keys of the given mapping (by key-value pairs
    or a ``**kwds`` dictionary), the keys are also added in lowercase form,
    and there is a backwards mapping from value to key.
    """

    def __init__(self, *args, **kwds):
        dict.__init__(self)
        self._keys = []
        for k, v in itertools.chain(kwds.items(), args):
            self[k] = v
            self[k.lower()] = v
            self[v] = k
            self._keys.append(k)

    def __getattr__(self, k):
        return self[k]



# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
#
# *****************************************************************************

"""Basic test suite for the scanner."""

from zapf.device import typecode_description
from zapf.scan import Scanner
from zapf.spec import v_2021_09


def test_scan(plc_io):
    scanner = Scanner(plc_io, plc_io.log)

    plc_data = scanner.get_plc_data()
    if plc_data.magicstr == '2015_02':
        assert plc_data.indexer_addr == 64
        assert plc_data.indexer_num_slots == 1
        assert plc_data.indexer_slot_size == 34
        assert plc_data.indexer_data_size == 34
        assert plc_data.indexer_size == 36
        assert plc_data.plc_name == 'lazy test plc'
        assert plc_data.plc_version == '0.0.1-alpha'
    else:
        assert plc_data.magicstr == '2021_09'
        assert plc_data.indexer_addr == 6
        assert (plc_data.indexer_num_slots,
                plc_data.indexer_slot_size) in ((3,64), (4,48))
        assert plc_data.indexer_data_size == \
            plc_data.indexer_num_slots * plc_data.indexer_slot_size
        assert plc_data.indexer_size == 194
        assert plc_data.plc_name == 'testplc_2021_09.py'
        assert plc_data.plc_version == \
            'https://forge.frm2.tum.de/review/mlz/pils/zapf:v2.1-alpha'
        assert plc_data.plc_description == 'simulation for testing zapf'
        assert isinstance(plc_data.query_descriptor(32767), v_2021_09.DebugDescriptor)
        assert plc_data.query_descriptor(32767).text == \
            'PLC_Problem 42, please call 137'
    assert plc_data.plc_author == 'anonymous\nauthor'

    for devinfo in scanner.scan_devices():
        # the test devices have the same name as the typecode description
        max_devname_len = plc_data.indexer_slot_size-21 \
            if plc_data.magicstr == '2021_09' else plc_data.indexer_data_size
        assert typecode_description(devinfo.typecode)[:max_devname_len] == devinfo.name
        # make sure get_info has all required keys in the info dict
        for key in ('description', 'aux', 'params', 'funcs', 'tables',
                    'flags', 'absmin', 'absmax', 'unit'):
            assert key in devinfo.info



import logging
import sys
import time

from zapf import io, scan

logging.basicConfig(level=0)

io = io.PlcIO(sys.argv[1], logging.root)
t0 = time.time()
scan = scan.Scanner(io, logging.root)
devs = list(scan.get_devices())
print(f'Readout complete in {time.time() - t0:.1f} s')
print('Running cache queries, Ctrl-C to abort')
io.start_cache()
while True:
    time.sleep(0.5)



#!/usr/bin/env python3
# pylint: disable=invalid-name
# *****************************************************************************
# PILS PLC client library
# Copyright (c) 2019-2021 by the authors, see LICENSE
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Module authors:
#   Georg Brandl <g.brandl@fz-juelich.de>
#
# *****************************************************************************

"""Run the integrated simulator standalone, without creating a Zapf proto."""

import argparse
import itertools
import sys
import threading
import time
from pathlib import Path

from zapf.simulator.runtime import Memory
from zapf.simulator.server import AdsServer, ModbusServer


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-a', action='store_true', help='Use ADS server')
    parser.add_argument('file', help='File with the PLC simulation', type=Path)
    args = parser.parse_args()

    # needs to be created before reading the PLC code!
    mem = Memory()

    code = args.file.read_text()
    globs = {}
    # pylint: disable=exec-used
    exec(code, globs)  # noqa: S102

    mainfunc = globs['Main']
    if not getattr(mainfunc, 'is_program', False):
        raise RuntimeError('main function must be a program')

    cond = threading.Condition()
    srv = AdsServer(mem, cond, 49000) if args.a else ModbusServer(mem, cond, 5002)
    threading.Thread(target=srv.serve_forever).start()

    print('Starting main PLC loop.')
    try:
        for i in itertools.count():
            if i % 100 == 0:
                print(f'\r{i:>10} cycles', end='')
                sys.stdout.flush()
            with cond:
                mainfunc()
                cond.notify()
            time.sleep(.005)
    except KeyboardInterrupt:
        srv.shutdown()
        sys.exit(0)


if __name__ == '__main__':
    main()
